from math import isnan

from pandas import concat, DataFrame, Series
from scipy.stats.stats import pearsonr

from bamboo.lib.utils import minint, parse_float


class Aggregation(object):
    """Abstract class for all aggregations.

    :param column: Column to aggregate.
    :param columns: List of columns to aggregate.
    :param formula_name: The string to refer to this aggregation.
    """
    column = None
    columns = None
    formula_name = None

    def __init__(self, name, groups, dframe):
        self.name = name
        self.groups = groups
        self.dframe = dframe

    def eval(self, columns):
        self.columns = columns
        self.column = columns[0] if len(columns) else None
        return self.group() if self.groups else self.agg()

    def group(self):
        """For when aggregation is called with a group parameter."""
        return self._groupby().agg(self.formula_name)

    def agg(self):
        """For when aggregation is called without a group parameter."""
        result = float(self.column.__getattribute__(self.formula_name)())
        return self._value_to_dframe(result)

    def _value_to_dframe(self, value):
        return DataFrame({self.name: Series([value])})

    def _groupby(self):
        return self.dframe[self.groups].join(concat(
            self.columns, axis=1)).groupby(self.groups, as_index=False)


class CountAggregation(Aggregation):
    """Calculate the count of rows fulfilling the criteria in the formula.

    N/A values are ignored unless there are no arguments to the function, in
    which case it returns the number of rows in the dataset.

    Written as ``count(CRITERIA)``. Where `CRITERIA` is an optional boolean
    expression that signifies which rows are to be counted.
    """
    formula_name = 'count'

    def group(self):
        if self.column is not None:
            joined = self.dframe[self.groups].join(
                self.column)
            joined = joined[self.column]
            groupby = joined.groupby(self.groups, as_index=False)
            result = groupby.agg(
                self.formula_name)[self.column.name].reset_index()
        else:
            result = self.dframe[self.groups]
            result = result.groupby(self.groups).apply(lambda x: len(x)).\
                reset_index().rename(columns={0: self.name})

        return result

    def agg(self):
        if self.column is not None:
            self.column = self.column[self.column]
            result = float(self.column.__getattribute__(self.formula_name)())
        else:
            result = len(self.dframe)

        return self._value_to_dframe(result)


class RatioAggregation(Aggregation):
    """Calculate the ratio.

    Columns with N/A for either the numerator or denominator are ignored.  This
    will store associated numerator and denominator columns.  Written as
    ``ratio(NUMERATOR, DENOMINATOR)``. Where `NUMERATOR` and `DENOMINATOR` are
    both valid formulas.
    """
    formula_name = 'ratio'

    def group(self):
        return self._group(self.columns)

    def _group(self, columns):
        dframe = self.dframe[self.groups]
        dframe = self._build_dframe(dframe, columns)
        groupby = dframe.groupby(self.groups, as_index=False)
        return self._add_calculated_column(groupby.sum())

    def agg(self):
        dframe = DataFrame(index=self.column.index)

        dframe = self._build_dframe(dframe, self.columns)
        column_names = [self.__name_for_idx(i) for i in xrange(0, 2)]
        dframe = dframe.dropna(subset=column_names)

        dframe = DataFrame([dframe.sum().to_dict()])

        return self._add_calculated_column(dframe)

    def reduce(self, dframe, columns):
        """Reduce the columns and store in `dframe`.

        :param dframe: The DataFrame to reduce.
        :param columns: Columns in the DataFrame to reduce on.
        """
        self.columns = columns
        self.column = columns[0]
        new_dframe = self.agg()

        for column in new_dframe.columns:
            dframe[column] += new_dframe[column]

        dframe[self.name] = self.__agg_dframe(dframe)

        return dframe

    def __name_for_idx(self, idx):
        return '%s_%s' % (self.name, {
            0: 'numerator',
            1: 'denominator',
        }[idx])

    def _build_dframe(self, dframe, columns):
        for idx, column in enumerate(columns):
            column.name = self.__name_for_idx(idx)

        return concat([dframe] + [DataFrame(col) for col in columns], axis=1)

    def _add_calculated_column(self, dframe):
        column = dframe[self.__name_for_idx(0)].apply(float) /\
            dframe[self.__name_for_idx(1)]
        column.name = self.name

        return dframe.join(column)

    def __agg_dframe(self, dframe):
        return dframe[self.__name_for_idx(0)].apply(float) /\
            dframe[self.__name_for_idx(1)]


class ArgMaxAggregation(Aggregation):
    """Return the index for the maximum of a column.

    Written as ``argmax(FORMULA)``. Where `FORMULA` is a valid formula.
    """
    formula_name = 'argmax'

    def group(self):
        """For when aggregation is called with a group parameter."""
        self.column = self.column.apply(lambda value: parse_float(value))
        group_dframe = self.dframe[self.groups].join(self.column)
        indices = group_dframe.reset_index().set_index(
            self.groups + [self.name])

        def max_index_for_row(row):
            groups = row[self.groups]
            value = row[self.name]

            xsection = indices.xs(groups, level=self.groups)

            if isnan(value):
                return minint()

            max_index = xsection.get_value(value, 'index')

            if isinstance(max_index, Series):
                max_index = max_index.max()

            return max_index

        groupby_max = self._groupby().max().reset_index()
        column = groupby_max.apply(max_index_for_row, axis=1).apply(int)
        column.name = self.name

        return DataFrame(column).join(groupby_max[self.groups])


class MaxAggregation(Aggregation):
    """Calculate the maximum.

    Written as ``max(FORMULA)``. Where `FORMULA` is a valid formula.
    """
    formula_name = 'max'


class MeanAggregation(RatioAggregation, Aggregation):
    """Calculate the arithmetic mean.

    Written as ``mean(FORMULA)``. Where `FORMULA` is a valid formula.

    Because mean is a ratio this inherits from `RatioAggregation` to
    use its generic reduce implementation.
    """
    formula_name = 'mean'

    def group(self):
        return self._group([self.column, Series([1] * len(self.column))])

    def agg(self):
        dframe = DataFrame(index=[0])

        columns = [
            Series([col]) for col in [self.column.sum(), len(self.column)]]

        dframe = self._build_dframe(dframe, columns)
        dframe = DataFrame([dframe.sum().to_dict()])

        return self._add_calculated_column(dframe)


class MedianAggregation(Aggregation):
    """Calculate the median. Written as ``median(FORMULA)``.

    Where `FORMULA` is a valid formula.
    """
    formula_name = 'median'


class MinAggregation(Aggregation):
    """Calculate the minimum.

    Written as ``min(FORMULA)``. Where `FORMULA` is a valid formula.
    """
    formula_name = 'min'


class NewestAggregation(Aggregation):
    """Return the second column's value at the newest row in the first column.

    Find the maximum value for the first column and return the entry at that
    row from the second column.

    Written as ``newest(INDEX_FORMULA, VALUE_FORMULA)`` where ``INDEX_FORMULA``
    and ``VALUE_FORMULA`` are valid formulae.
    """
    formula_name = 'newest'
    value_column = 1

    def agg(self):
        index, values = self.columns
        index.name = 'index'
        values.name = self.name

        idframe = DataFrame(values).join(index).dropna().reset_index()
        idx = idframe['index'].argmax()
        result = idframe[self.name].get_value(idx)

        return self._value_to_dframe(result)

    def group(self):
        argmax_agg = ArgMaxAggregation(self.name, self.groups, self.dframe)
        argmax_df = argmax_agg.eval(self.columns)
        indices = argmax_df.pop(self.name)

        newest_col = self.columns[self.value_column][indices]
        newest_col.index = argmax_df.index
        newest_col.name = self.name

        return argmax_df.join(newest_col)


class PearsonAggregation(Aggregation):
    """Calculate the Pearson correlation and associatd p-value.

    Calculate the Pearson correlation coefficient between two columns and the
    p-value for that correlation coefficient.

    Written as ``pearson(FORMULA1, FORMULA2)``. Where ``FORMULA1`` and
    ``FORMULA2`` are valid formulae.
    """
    formula_name = 'pearson'

    def agg(self):
        coor, pvalue = self.__pearsonr(self.columns)
        pvalue_column = Series([pvalue], name=self.__pvalue_name)
        return self._value_to_dframe(coor).join(pvalue_column)

    def group(self):
        def pearson(dframe):
            columns = [dframe[name] for name in dframe.columns[-2:]]

            return DataFrame([self.__pearsonr(columns)],
                             columns=[self.name, self.__pvalue_name])

        groupby = self._groupby()
        dframe = groupby.apply(pearson).reset_index()

        # remove extra index column
        del dframe[dframe.columns[len(self.groups)]]
        return dframe

    def __pearsonr(self, columns):
        columns = [c.dropna() for c in columns]
        shared_index = reduce(
            lambda x, y: x.index.intersection(y.index), columns)
        columns = [c.ix[shared_index] for c in columns]

        return pearsonr(*columns)

    @property
    def __pvalue_name(self):
        return '%s_pvalue' % self.name


class StandardDeviationAggregation(Aggregation):
    """Calculate the standard deviation. Written as ``std(FORMULA)``.

    Where `FORMULA` is a valid formula.
    """
    formula_name = 'std'


class SumAggregation(Aggregation):
    """Calculate the sum.

    Written as ``sum(FORMULA)``. Where `FORMULA` is a valid formula.
    """
    formula_name = 'sum'

    def reduce(self, dframe, columns):
        self.columns = columns
        self.column = columns[0]
        dframe[self.name] += self.agg()[self.name]

        return dframe


class VarianceAggregation(Aggregation):
    """Calculate the variance. Written as ``var(FORMULA)``.

    Where `FORMULA` is a valid formula.
    """
    formula_name = 'var'


# dict of formula names to aggregation classes
AGGREGATIONS = {
    cls.formula_name: cls for cls in
    Aggregation.__subclasses__()
    if cls.formula_name
}



from base64 import b64encode
import simplejson as json
import re

from bson import json_util


# MongoDB keys
MONGO_RESERVED_KEYS = ['_id']
MONGO_RESERVED_KEY_PREFIX = 'MONGO_RESERVED_KEY'
MONGO_RESERVED_KEY_STRS = [MONGO_RESERVED_KEY_PREFIX + key
                           for key in MONGO_RESERVED_KEYS]
ILLEGAL_VALUES = ['$', '.']
REPLACEMENT_VALUES = [b64encode(value) for value in ILLEGAL_VALUES]

RE_ILLEGAL_MAP = [(re.compile(r'\%s' % value), REPLACEMENT_VALUES[idx]) for
                  idx, value in enumerate(ILLEGAL_VALUES)]
RE_LEGAL_MAP = [(re.compile(r'\%s' % value), ILLEGAL_VALUES[idx]) for
                idx, value in enumerate(REPLACEMENT_VALUES)]


def dump_mongo_json(obj):
    """Dump JSON using BSON conversion.

    Args:

    :param obj: Datastructure to dump as JSON.

    :returns: JSON string of dumped `obj`.
    """
    return json.dumps(obj, default=json_util.default)


def remove_mongo_reserved_keys(_dict):
    """Remove any keys reserved for MongoDB from `_dict`.

    Check for `MONGO_RESERVED_KEYS` in stored dictionary.  If found replace
    with unprefixed, if not found remove reserved key from dictionary.

    Args:

    :param _dict: Dictionary to remove reserved keys from.

    :returns: Dictionary with reserved keys removed.
    """
    for key in MONGO_RESERVED_KEYS:
        prefixed_key = mongo_prefix_reserved_key(key)
        if _dict.get(prefixed_key):
            _dict[key] = _dict.pop(prefixed_key)
        else:
            # remove mongo reserved keys
            del _dict[key]

    return _dict


def mongo_prefix_reserved_key(key, prefix=MONGO_RESERVED_KEY_PREFIX):
    """Prefix reserved key."""
    return '%s%s' % (prefix, key)


def reserve_encoded(string):
    """Return encoding prefixed string."""
    return mongo_prefix_reserved_key(string) if\
        string in MONGO_RESERVED_KEYS else string


def dict_from_mongo(_dict):
    for key, value in _dict.items():
        if isinstance(value, list):
            value = [dict_from_mongo(obj)
                     if isinstance(obj, dict) else obj for obj in value]
        elif isinstance(value, dict):
            value = dict_from_mongo(value)

        if _was_encoded_for_mongo(key):
            del _dict[key]
            _dict[_decode_from_mongo(key)] = value

    return _dict


def dict_for_mongo(_dict):
    """Encode all keys in `_dict` for MongoDB."""
    for key, value in _dict.items():
        if _is_invalid_for_mongo(key):
            del _dict[key]
            key = _encode_for_mongo(key)

        if isinstance(value, list):
            _dict[key] = [dict_for_mongo(obj) if isinstance(obj, dict) else obj
                          for obj in value]
        elif isinstance(value, dict):
            _dict[key] = dict_for_mongo(value)
        else:
            _dict[key] = value

    return _dict


def _decode_from_mongo(key):
    return reduce(lambda s, expr: expr[0].sub(expr[1], s),
                  RE_LEGAL_MAP, key)


def _encode_for_mongo(key):
    """Encode illegal MongoDB characters in string.

    Base64 encode any characters in a string that cannot be MongoDB keys. This
    includes any '$' and any '.'. '$' are supposed to be allowed as the
    non-first character but the current version of MongoDB does not allow any
    occurence of '$'.

    :param key: The string to remove characters from.

    :returns: The string with illegal keys encoded.
    """
    return reduce(lambda s, expr: expr[0].sub(expr[1], s),
                  RE_ILLEGAL_MAP, key)


def _is_invalid_for_mongo(key):
    """Return if string is invalid for storage in MongoDB."""
    return any([key.count(value) > 0 for value in ILLEGAL_VALUES])


def _was_encoded_for_mongo(key):
    return any([key.count(value) > 0 for value in REPLACEMENT_VALUES])



from math import ceil

from pymongo.errors import AutoReconnect

from bamboo.config.db import Database
from bamboo.core.frame import BAMBOO_RESERVED_KEYS
from bamboo.lib.decorators import classproperty
from bamboo.lib.mongo import dict_for_mongo, remove_mongo_reserved_keys


class AbstractModel(object):
    """An abstact class for all MongoDB models.

    Attributes:

    - __collection__: The MongoDB collection to communicate with.
    - STATE: A key with which to store state.
    - STATE_PENDING: A value for the pending state.
    - STATE_READY: A value for the ready state.

    """

    __collection__ = None

    # delimiter when passing multiple groups as a string
    GROUP_DELIMITER = ','
    ERROR_MESSAGE = 'error_message'

    DB_READ_BATCH_SIZE = 1000
    DB_SAVE_BATCH_SIZE = 2000
    MIN_BATCH_SIZE = 50

    STATE = 'state'
    STATE_FAILED = 'failed'
    STATE_PENDING = 'pending'
    STATE_READY = 'ready'

    @property
    def state(self):
        return self.record[self.STATE]

    @property
    def error_message(self):
        return self.record.get(self.ERROR_MESSAGE)

    @property
    def is_pending(self):
        return self.state == self.STATE_PENDING

    @property
    def is_ready(self):
        return self.state == self.STATE_READY

    @property
    def record_ready(self):
        return self.record is not None and self.state == self.STATE_READY

    @property
    def clean_record(self):
        """Remove reserved keys from records."""
        _dict = {
            key: value for (key, value) in self.record.items() if not key in
            BAMBOO_RESERVED_KEYS
        }
        return remove_mongo_reserved_keys(_dict)

    @classmethod
    def set_collection(cls, collection_name):
        """Return a MongoDB collection for the passed name.

        :param collection_name: The name of collection to return.

        :returns: A MongoDB collection from the current database.
        """
        return Database.db()[collection_name]

    @classproperty
    @classmethod
    def collection(cls):
        """Set the internal collection to the class' collection name."""
        if not cls.__collection__:
            cls.__collection__ = AbstractModel.set_collection(
                cls.__collectionname__)

        return cls.__collection__

    @classmethod
    def create(cls, *args):
        model = cls()
        return model.save(*args)

    @classmethod
    def find(cls, query_args, as_dict=False, as_cursor=False):
        """An interface to MongoDB's find functionality.

        :param query_args: An optional QueryArgs to hold the query arguments.
        :param as_cursor: If true, return the cursor.
        :param as_dict: If true, return dicts and not model instances.

        :returns: A list of dicts or model instances for each row returned.
        """
        cursor = cls.collection.find(query_args.query,
                                     query_args.select,
                                     sort=query_args.order_by,
                                     limit=query_args.limit)

        if as_cursor:
            return cursor
        else:
            return [record for record in cursor] if as_dict else [
                cls(record) for record in cursor
            ]

    @classmethod
    def find_one(cls, query, select=None):
        """Return the first row matching `query` and `select` from MongoDB.

        :param query: A query to pass to MongoDB.
        :param select: An optional select to pass to MongoDB.

        :returns: A model instance of the row returned for this query and
            select.
        """
        return cls(cls.collection.find_one(query, select))

    @classmethod
    def batch_save(cls, dframe):
        """Save records in batches to avoid document size maximum setting.

        :param dframe: A DataFrame to save in the current model.
        """
        def command(records):
            cls.collection.insert(records)

        batch_size = cls.DB_SAVE_BATCH_SIZE

        cls.__batch_command_wrapper(command, dframe, batch_size)

    @classmethod
    def __batch_command_wrapper(cls, command, dframe, batch_size):
        try:
            cls.__batch_command(command, dframe, batch_size)
        except AutoReconnect:
            batch_size /= 2

            # If batch size drop is less than MIN_BATCH_SIZE, assume the
            # records are too large or there is another error and fail.
            if batch_size >= cls.MIN_BATCH_SIZE:
                cls.__batch_command_wrapper(command, dframe, batch_size)

    @classmethod
    def __batch_command(cls, command, dframe, batch_size):
        batches = int(ceil(float(len(dframe)) / batch_size))

        for batch in xrange(0, batches):
            start = batch * batch_size
            end = start + batch_size
            records = [
                row.to_dict() for (_, row) in dframe[start:end].iterrows()]
            command(records)

    def __init__(self, record=None):
        """Instantiate with data in `record`."""
        self.record = record

    def __nonzero__(self):
        return self.record is not None

    def failed(self, message=None):
        """Perist the state of the current instance to `STATE_FAILED`.

        :params message: A string store as the error message, default None.
        """
        doc = {self.STATE: self.STATE_FAILED}

        if message:
            doc.update({self.ERROR_MESSAGE: message})

        self.update(doc)

    def pending(self):
        """Perist the state of the current instance to `STATE_PENDING`"""
        self.update({self.STATE: self.STATE_PENDING})

    def ready(self):
        """Perist the state of the current instance to `STATE_READY`"""
        self.update({self.STATE: self.STATE_READY})

    def delete(self, query):
        """Delete rows matching query.

        :param query: The query for rows to delete.
        """
        self.collection.remove(query)

    def save(self, record):
        """Save `record` in this model's collection.

        Save the record in the model instance's collection and set the internal
        record of this instance to the passed in record.

        :param record: The dict to save in the model's collection.

        :returns: The record passed in.
        """
        self.collection.insert(record)
        self.record = record

        return self

    def update(self, record):
        """Update the current instance with `record`.

        Update the current model instance based on its `_id`, set it to the
        passed in `record`.

        :param record: The record to replace the instance's data with.
        """
        record = dict_for_mongo(record)
        id_dict = {'_id': self.record['_id']}
        self.collection.update(id_dict, {'$set': record})

        # Set record to the latest record from the database
        self.record = self.__class__.collection.find_one(id_dict)

    def split_groups(self, groups):
        """Split a string based on the group delimiter"""
        return groups.split(self.GROUP_DELIMITER) if groups else []

    def join_groups(self, groups):
        return self.GROUP_DELIMITER.join(groups)



from functools import partial
import simplejson as json
import os
import tempfile

from celery.exceptions import RetryTaskError
from celery.task import task
import pandas as pd

from bamboo.core.frame import BambooFrame
from bamboo.lib.async import call_async


@task(ignore_result=True)
def import_dataset(dataset, file_reader, delete=False):
    """For reading a URL and saving the corresponding dataset.

    Import a DataFrame using the provided `file_reader` function. All
    exceptions are caught and on exception the dataset is marked as failed and
    set for deletion after 24 hours.

    :param dataset: The dataset to import into.
    :param file_reader: Function for reading the dataset.
    :param delete: Delete filepath_or_buffer after import, default False.
    """
    try:
        dframe = file_reader()
        dataset.save_observations(dframe)
    except Exception as e:
        if isinstance(e, RetryTaskError):
            raise e
        else:
            dataset.failed(e.__str__())
            dataset.delete(countdown=86400)


def csv_file_reader(name, na_values=[], delete=False):
    try:
        return BambooFrame(pd.read_csv(
            name, encoding='utf-8', na_values=na_values)).recognize_dates()
            #pd.read_csv(name, encoding='utf-8')).recognize_dates()
    finally:
        if delete:
            os.unlink(name)


def json_file_reader(content):
    return BambooFrame(json.loads(content)).recognize_dates()


class ImportableDataset(object):
    def import_from_url(self, url, na_values=[], allow_local_file=False):
        """Load a URL, read from a CSV, add data to dataset.

        :param url: URL to load file from.
        :param allow_local_file: Allow URL to refer to a local file.

        :raises: `IOError` for an unreadable file or a bad URL.

        :returns: The created dataset.
        """
        if not allow_local_file and isinstance(url, basestring)\
                and url[0:4] == 'file':
            raise IOError

        call_async(
            import_dataset, self, partial(
                csv_file_reader, url, na_values=na_values))

        return self

    def import_from_csv(self, csv_file):
        """Import data from a CSV file.

        .. note::

            Write to a named tempfile in order  to get a handle for pandas'
            `read_csv` function.

        :param csv_file: The CSV File to create a dataset from.

        :returns: The created dataset.
        """
        if 'file' in dir(csv_file):
            csv_file = csv_file.file

        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        tmpfile.write(csv_file.read())

        # pandas needs a closed file for *read_csv*
        tmpfile.close()

        call_async(import_dataset, self, partial(
            csv_file_reader, tmpfile.name, delete=True))

        return self

    def import_from_json(self, json_file):
        """Impor data from a JSON file.

        :param json_file: JSON file to import.
        """
        content = json_file.file.read()

        call_async(import_dataset, self,
                   partial(json_file_reader, content))

        return self

    def import_schema(self, schema):
        """Create a dataset from a SDF schema file (JSON).

        :param schema: The SDF (JSON) file to create a dataset from.

        :returns: The created dataset.
        """
        try:
            schema = json.loads(schema.file.read())
        except AttributeError:
            schema = json.loads(schema)

        self.set_schema(schema)
        self.ready()

        return self



from time import sleep

from pandas import concat
import simplejson as json

from bamboo.core.frame import BAMBOO_RESERVED_KEYS, PARENT_DATASET_ID
from bamboo.lib.mongo import MONGO_RESERVED_KEY_STRS
from bamboo.models.dataset import Dataset
from bamboo.tests.controllers.test_abstract_datasets import\
    TestAbstractDatasets
from bamboo.tests.decorators import requires_async


class TestDatasetsMerge(TestAbstractDatasets):

    def _post_merge(self, dataset_ids):
        dataset_id1, dataset_id2 = dataset_ids
        return json.loads(self.controller.merge(
            dataset_ids=json.dumps(dataset_ids),
            mapping=json.dumps({
                dataset_id1: {
                    "food_type": "food_type_2",
                },
                dataset_id2: {
                    "code": "comments",
                    "food_type": "food_type_2",
                },
            })))[Dataset.ID]

    @requires_async
    def test_merge_datasets_0_not_enough(self):
        result = json.loads(self.controller.merge(dataset_ids=json.dumps([])))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(self.controller.ERROR in result)

    @requires_async
    def test_merge_datasets_1_not_enough(self):
        dataset_id = self._post_file()
        result = json.loads(self.controller.merge(
            dataset_ids=json.dumps([dataset_id])))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(self.controller.ERROR in result)

    @requires_async
    def test_merge_datasets_must_exist(self):
        dataset_id = self._post_file()
        result = json.loads(self.controller.merge(
            dataset_ids=json.dumps([dataset_id, 0000])))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(self.controller.ERROR in result)

    def test_merge_datasets(self):
        dataset_id1 = self._post_file()
        dataset_id2 = self._post_file()
        result = json.loads(self.controller.merge(
            dataset_ids=json.dumps([dataset_id1, dataset_id2])))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Dataset.ID in result)

        datasets = [Dataset.find_one(dataset_id)
                    for dataset_id in [dataset_id1, dataset_id2]]

        for dataset in datasets:
            self.assertTrue(result[Dataset.ID] in dataset.merged_dataset_ids)

        dframe1 = datasets[0].dframe()
        merged_dataset = Dataset.find_one(result[Dataset.ID])
        merged_rows = merged_dataset.observations()

        for row in merged_rows:
            self.assertTrue(PARENT_DATASET_ID in row.keys())

        merged_dframe = merged_dataset.dframe()

        self.assertEqual(len(merged_dframe), 2 * len(dframe1))

        expected_dframe = concat([dframe1, dframe1],
                                 ignore_index=True)

        self.assertEqual(list(merged_dframe.columns),
                         list(expected_dframe.columns))

        self._check_dframes_are_equal(merged_dframe, expected_dframe)

    @requires_async
    def test_merge_datasets_async(self):
        dataset_id1 = self._post_file()
        dataset_id2 = self._post_file()

        self.assertEqual(
            Dataset.find_one(dataset_id1).state,
            Dataset.STATE_PENDING)
        self.assertEqual(
            Dataset.find_one(dataset_id2).state,
            Dataset.STATE_PENDING)

        result = json.loads(self.controller.merge(
            dataset_ids=json.dumps([dataset_id1, dataset_id2])))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Dataset.ID in result)

        merged_id = result[Dataset.ID]

        while True:
            datasets = [Dataset.find_one(dataset_id)
                        for dataset_id in [merged_id, dataset_id1, dataset_id2]
                        ]

            if all([dataset.record_ready for dataset in datasets]) and all(
                    [d.merged_dataset_ids for d in datasets[1:]]):
                break

            sleep(self.SLEEP_DELAY)

        datasets = [Dataset.find_one(dataset_id)
                    for dataset_id in [dataset_id1, dataset_id2]]

        for dataset in datasets:
            self.assertTrue(merged_id in dataset.merged_dataset_ids)

        dframe1 = datasets[0].dframe()
        merged_dataset = Dataset.find_one(merged_id)
        merged_rows = merged_dataset.observations()

        for row in merged_rows:
            self.assertTrue(PARENT_DATASET_ID in row.keys())

        merged_dframe = merged_dataset.dframe()

        self.assertEqual(len(merged_dframe), 2 * len(dframe1))

        expected_dframe = concat([dframe1, dframe1],
                                 ignore_index=True)

        self.assertEqual(list(merged_dframe.columns),
                         list(expected_dframe.columns))

        self._check_dframes_are_equal(merged_dframe, expected_dframe)

    @requires_async
    def test_merge_datasets_add_calc_async(self):
        dataset_id1 = self._post_file('good_eats_large.csv')
        dataset_id2 = self._post_file('good_eats_large.csv')
        result = json.loads(self.controller.merge(
            dataset_ids=json.dumps([dataset_id1, dataset_id2])))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Dataset.ID in result)

        self.dataset_id = result[Dataset.ID]
        self.schema = json.loads(
            self.controller.info(self.dataset_id))[Dataset.SCHEMA]

        self._post_calculations(['amount < 4'])

    def test_merge_datasets_no_reserved_keys(self):
        dataset_id1 = self._post_file()
        dataset_id2 = self._post_file()
        result = json.loads(self.controller.merge(
            dataset_ids=json.dumps([dataset_id1, dataset_id2])))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Dataset.ID in result)

        response = json.loads(self.controller.show(result[Dataset.ID]))
        row_keys = sum([row.keys() for row in response], [])

        for reserved_key in BAMBOO_RESERVED_KEYS + MONGO_RESERVED_KEY_STRS:
            self.assertFalse(reserved_key in row_keys)

    def test_merge_with_map(self):
        dataset_id1 = self._post_file()
        dataset_id2 = self._post_file('good_eats_aux.csv')
        merged_dataset_id = self._post_merge([dataset_id1, dataset_id2])

        expected_columns = Dataset.find_one(
            dataset_id1).dframe().columns.tolist()
        expected_columns.remove("food_type")
        expected_columns.append("food_type_2")
        expected_columns = set(expected_columns)

        merged_dataset = Dataset.find_one(merged_dataset_id)
        new_columns = set(merged_dataset.dframe().columns)

        self.assertEquals(expected_columns, new_columns)

    def test_merge_with_map_update(self):
        dataset_id1 = self._post_file()
        dataset_id2 = self._post_file('good_eats_aux.csv')
        merged_dataset_id = self._post_merge([dataset_id1, dataset_id2])

        original_ds2 = json.loads(self.controller.show(dataset_id2))
        original_length = len(original_ds2)
        original_merge = json.loads(self.controller.show(merged_dataset_id))
        original_merge_length = len(original_merge)

        self._put_row_updates(dataset_id2, 'good_eats_aux_update.json')
        response = json.loads(self.controller.show(dataset_id2))
        new_length = len(response)

        for new_row in response:
            if new_row not in original_ds2:
                break

        response = json.loads(self.controller.show(merged_dataset_id))

        for new_merge_row in response:
            if new_merge_row not in original_merge:
                break

        new_merge_length = len(response)

        self.assertEqual(original_length + 1, new_length)
        self.assertEqual(original_merge_length + 1, new_merge_length)
        self.assertEqual(new_row['food_type'], new_merge_row['food_type_2'])
        self.assertEqual(new_row['code'], new_merge_row['comments'])
        merged_dataset = Dataset.find_one(merged_dataset_id)



# future must be first
from __future__ import division
import operator

import numpy as np
from scipy.stats import percentileofscore

from bamboo.lib.datetools import safe_parse_date_to_unix_time,\
    parse_str_to_unix_time


class EvalTerm(object):
    """Base class for evaluation."""

    def __init__(self, tokens):
        self.tokens = tokens
        self.value = tokens[0]

    def operator_operands(self, tokenlist):
        """Generator to extract operators and operands in pairs."""
        it = iter(tokenlist)

        while 1:
            try:
                yield (it.next(), it.next())
            except StopIteration:
                break

    def operation(self, oper, result, val):
        return self.operations[oper](result, val)


class EvalConstant(EvalTerm):
    """Class to evaluate a parsed constant or variable."""

    def eval(self, row, context):
        try:
            return np.float64(self.value)
        except ValueError:
            # it may be a variable
            field = self.field(row)
            if field:
                # it is a variable, save as dependency
                context.dependent_columns.add(self.value)

            # test is date and parse as date
            return self.__parse_field(field, context)

    def __parse_field(self, field, context):
            schema = context.schema

            if schema and schema.is_date_simpletype(self.value):
                field = safe_parse_date_to_unix_time(field)

            return field

    def field(self, row):
        return row.get(self.value)


class EvalString(EvalTerm):
    """Class to evaluate a parsed string."""

    def eval(self, row, context):
        return self.value


class EvalSignOp(EvalTerm):
    """Class to evaluate expressions with a leading + or - sign."""

    def __init__(self, tokens):
        self.sign, self.value = tokens[0]

    def eval(self, row, context):
        mult = {'+': 1, '-': -1}[self.sign]
        return mult * self.value.eval(row, context)


class EvalBinaryArithOp(EvalTerm):
    """Class for evaluating binary arithmetic operations."""

    operations = {
        '+': operator.__add__,
        '-': operator.__sub__,
        '*': operator.__mul__,
        '/': operator.__truediv__,
        '^': operator.__pow__,
    }

    def eval(self, row, context):
        result = np.float64(self.value[0].eval(row, context))

        for oper, val in self.operator_operands(self.value[1:]):
            val = np.float64(val.eval(row, context))
            result = self.operation(oper, result, val)
            if np.isinf(result):
                return np.nan

        return result


class EvalMultOp(EvalBinaryArithOp):
    """Class to distinguish precedence of multiplication/division expressions.
    """
    pass


class EvalPlusOp(EvalBinaryArithOp):
    """Class to distinguish precedence of addition/subtraction expressions.
    """
    pass


class EvalExpOp(EvalBinaryArithOp):
    """Class to distinguish precedence of exponentiation expressions.
    """
    pass


class EvalComparisonOp(EvalTerm):
    """Class to evaluate comparison expressions."""

    op_map = {
        "<": lambda a, b: a < b,
        "<=": lambda a, b: a <= b,
        ">": lambda a, b: a > b,
        ">=": lambda a, b: a >= b,
        "!=": lambda a, b: a != b,
        "==": lambda a, b: a == b,
    }

    def eval(self, row, context):
        val1 = np.float64(self.value[0].eval(row, context))

        for oper, val in self.operator_operands(self.value[1:]):
            fn = EvalComparisonOp.op_map[oper]
            val2 = np.float64(val.eval(row, context))
            if not fn(val1, val2):
                break
            val1 = val2
        else:
            return True

        return False


class EvalNotOp(EvalTerm):
    """Class to evaluate not expressions."""

    def __init__(self, tokens):
        self.value = tokens[0][1]

    def eval(self, row, context):
        return not self.value.eval(row, context)


class EvalBinaryBooleanOp(EvalTerm):
    """Class for evaluating binary boolean operations."""

    operations = {
        'and': lambda p, q: p and q,
        'or': lambda p, q: p or q,
    }

    def eval(self, row, context):
        result = np.bool_(self.value[0].eval(row, context))

        for oper, val in self.operator_operands(self.value[1:]):
            val = np.bool_(val.eval(row, context))
            result = self.operation(oper, result, val)

        return result


class EvalAndOp(EvalBinaryBooleanOp):
    """Class to distinguish precedence of and expressions."""
    pass


class EvalOrOp(EvalBinaryBooleanOp):
    """Class to distinguish precedence of or expressions."""
    pass


class EvalInOp(EvalTerm):
    """Class to eval in expressions."""

    def eval(self, row, context):
        val_to_test = str(self.value[0].eval(row, context))
        val_list = []

        # loop through values, skip atom literal
        for val in self.value[1:]:
            val_list.append(val.eval(row, context))

        return val_to_test in val_list


class EvalCaseOp(EvalTerm):
    """Class to eval case statements."""

    def eval(self, row, context):
        # skip 'case' literal
        for token in self.value:
            case_result = token.eval(row, context)
            if case_result:
                return case_result

        return np.nan


class EvalMapOp(EvalTerm):
    """Class to eval map statements."""

    def eval(self, row, context):
        if self.tokens[0] == 'default' or self.tokens[0].eval(row, context):
            return self.tokens[1].eval(row, context)

        return False


class EvalFunction(object):
    """Class to eval functions."""

    def __init__(self, tokens):
        self.value = tokens[0][1]


class EvalDate(EvalFunction):
    """Class to evaluate date expressions."""

    def eval(self, row, context):
        # parse date from string
        return parse_str_to_unix_time(self.value.eval(row, context))


class EvalPercentile(EvalFunction):
    """Class to evaluate percentile expressions."""

    def eval(self, row, context):
        # parse date from string
        column = context.dframe[self.value.value]
        field = self.value.field(row)
        return percentileofscore(column, field)



import simplejson as json

from bamboo.lib.schema_builder import SIMPLETYPE
from bamboo.models.dataset import Dataset
from bamboo.tests.controllers.test_abstract_datasets import\
    TestAbstractDatasets
from bamboo.tests.mock import MockUploadedFile


class TestDatasetsFromSchema(TestAbstractDatasets):

    def setUp(self):
        TestAbstractDatasets.setUp(self)

    def _upload_from_schema(self, file_name):
        schema = open('tests/fixtures/%s' % file_name)
        mock_uploaded_file = MockUploadedFile(schema)

        return json.loads(self.controller.create(schema=mock_uploaded_file))

    def _upload_good_eats_schema(self):
        result = self._upload_from_schema('good_eats.schema.json')
        self._test_summary_built(result)

    def test_create_from_schema(self):
        self._upload_good_eats_schema()
        results = json.loads(self.controller.show(self.dataset_id))

        self.assertTrue(isinstance(results, list))
        self.assertEqual(len(results), 0)

        results = json.loads(self.controller.info(self.dataset_id))

        self.assertTrue(isinstance(results, dict))
        self.assertTrue(Dataset.SCHEMA in results.keys())
        self.assertTrue(Dataset.NUM_ROWS in results.keys())
        self.assertEqual(results[Dataset.NUM_ROWS], 0)
        self.assertTrue(Dataset.NUM_COLUMNS in results.keys())
        self.assertEqual(results[Dataset.NUM_COLUMNS], self.NUM_COLS)

    def test_create_from_schema_and_update(self):
        self._upload_good_eats_schema()
        results = json.loads(self.controller.show(self.dataset_id))

        self.assertFalse(len(results))

        dataset = Dataset.find_one(self.dataset_id)

        self.assertEqual(dataset.num_rows, 0)

        old_schema = dataset.schema
        self._put_row_updates()
        results = json.loads(self.controller.show(self.dataset_id))

        self.assertTrue(len(results))

        for result in results:
            self.assertTrue(isinstance(result, dict))
            self.assertTrue(len(result.keys()))

        dataset = Dataset.find_one(self.dataset_id)

        self.assertEqual(dataset.num_rows, 1)

        new_schema = dataset.schema

        self.assertEqual(set(old_schema.keys()), set(new_schema.keys()))
        for column in new_schema.keys():
            if new_schema.cardinality(column):
                self.assertEqual(new_schema.cardinality(column), 1)

    def test_create_one_from_schema_and_join(self):
        self._upload_good_eats_schema()
        left_dataset_id = self.dataset_id
        right_dataset_id = self._post_file('good_eats_aux.csv')

        on = 'food_type'
        dataset_id_tuples = [
            (left_dataset_id, right_dataset_id),
            (right_dataset_id, left_dataset_id),
        ]

        for dataset_ids in dataset_id_tuples:
            result = json.loads(self.controller.join(*dataset_ids, on=on))
            expected_schema_keys = set(sum([
                Dataset.find_one(dataset_id).schema.keys()
                for dataset_id in dataset_ids], []))

            self.assertTrue(isinstance(result, dict))
            self.assertTrue(Dataset.ID in result)
            merge_dataset_id = result[Dataset.ID]
            dataset = Dataset.find_one(merge_dataset_id)
            self.assertEqual(dataset.num_rows, 0)
            self.assertEqual(dataset.num_columns, len(expected_schema_keys))
            schema_keys = set(dataset.schema.keys())
            self.assertEqual(schema_keys, expected_schema_keys)

    def test_create_two_from_schema_and_join(self):
        self._upload_good_eats_schema()
        left_dataset_id = self.dataset_id

        schema = open('tests/fixtures/good_eats_aux.schema.json')
        mock_uploaded_file = MockUploadedFile(schema)
        result = json.loads(
            self.controller.create(schema=mock_uploaded_file))
        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Dataset.ID in result)
        right_dataset_id = result[Dataset.ID]

        on = 'food_type'
        dataset_id_tuples = [
            (left_dataset_id, right_dataset_id),
            (right_dataset_id, left_dataset_id),
        ]

        for dataset_ids in dataset_id_tuples:
            result = json.loads(self.controller.join(*dataset_ids, on=on))
            expected_schema_keys = set(sum([
                Dataset.find_one(dataset_id).schema.keys()
                for dataset_id in dataset_ids], []))

            self.assertTrue(isinstance(result, dict))
            self.assertTrue(Dataset.ID in result)
            merge_dataset_id = result[Dataset.ID]
            dataset = Dataset.find_one(merge_dataset_id)
            self.assertEqual(dataset.num_rows, 0)
            self.assertEqual(dataset.num_columns, len(expected_schema_keys))
            schema_keys = set(dataset.schema.keys())
            self.assertEqual(schema_keys, expected_schema_keys)

    def test_create_two_from_schema_and_join_and_update_vacuous_rhs(self):
        self._upload_good_eats_schema()
        left_dataset_id = self.dataset_id

        schema = open('tests/fixtures/good_eats_aux.schema.json')
        mock_uploaded_file = MockUploadedFile(schema)
        result = json.loads(
            self.controller.create(schema=mock_uploaded_file))
        right_dataset_id = result[Dataset.ID]

        on = 'food_type'
        result = json.loads(self.controller.join(
            left_dataset_id, right_dataset_id, on=on))
        merged_dataset_id = self.dataset_id = result[Dataset.ID]

        num_rows = 0
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

        num_rows += 1
        self._put_row_updates()
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

        num_rows += 1
        self._put_row_updates(left_dataset_id)
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

        self._put_row_updates(right_dataset_id)
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

    def test_create_two_from_schema_and_join_and_update_child_rhs(self):
        self._upload_good_eats_schema()
        left_dataset_id = self.dataset_id

        result = self._upload_from_schema('good_eats_aux.schema.json')
        right_dataset_id = result[Dataset.ID]

        on = 'food_type'
        result = json.loads(self.controller.join(
            left_dataset_id, right_dataset_id, on=on))
        merged_dataset_id = self.dataset_id = result[Dataset.ID]

        num_rows = 0
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

        num_rows += 1
        self._put_row_updates(file_name='good_eats_update_bg.json')
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

        self._put_row_updates(right_dataset_id,
                              file_name='good_eats_aux_update.json')
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

    def test_create_two_from_schema_and_join_and_update_lhs_rhs(self):
        self._upload_good_eats_schema()
        left_dataset_id = self.dataset_id

        result = self._upload_from_schema('good_eats_aux.schema.json')
        right_dataset_id = result[Dataset.ID]

        on = 'food_type'
        result = json.loads(self.controller.join(
            left_dataset_id, right_dataset_id, on=on))
        merged_dataset_id = self.dataset_id = result[Dataset.ID]

        num_rows = 0
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

        num_rows += 1
        self._put_row_updates(left_dataset_id,
                              file_name='good_eats_update_bg.json')
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))

        prev_num_cols = len(
            Dataset.find_one(merged_dataset_id).dframe().columns)
        self.dataset_id = right_dataset_id
        self._put_row_updates(right_dataset_id,
                              file_name='good_eats_aux_update.json')
        results = json.loads(self.controller.show(merged_dataset_id))
        self.assertEqual(num_rows, len(results))
        result = results[0]
        self.assertTrue('code' in result.keys())
        self.assertFalse(result['code'] is None)

    def test_schema_with_boolean_column(self):
        mock_csv_file = self._file_mock('wp_data.csv', add_prefix=True)
        mock_schema_file = self._file_mock(
            'wp_data.schema.json', add_prefix=True)
        result = json.loads(self.controller.create(csv_file=mock_csv_file,
                                                   schema=mock_schema_file))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Dataset.ID in result)

        dataset = Dataset.find_one(result[Dataset.ID])
        mock_schema_file.file.seek(0)

        for column, schema in json.loads(mock_schema_file.file.read()).items():
            self.assertEqual(schema[SIMPLETYPE],
                             dataset.schema[column][SIMPLETYPE])



from bamboo.core.parser import Parser
from bamboo.core.calculator import Calculator
from bamboo.lib.datetools import recognize_dates
from bamboo.models.calculation import Calculation
from bamboo.models.dataset import Dataset
from bamboo.tests.test_base import TestBase


class TestCalculator(TestBase):

    def setUp(self):
        TestBase.setUp(self)
        self.dataset = Dataset()
        self.dataset.save(
            self.test_dataset_ids['good_eats_with_calculations.csv'])
        dframe = recognize_dates(
            self.get_data('good_eats_with_calculations.csv'))
        self.dataset.save_observations(dframe)
        self.group = None
        self.parser = Parser(self.dataset)
        self.places = 5

    def _equal_msg(self, calculated, stored, formula):
        return '(calculated %s) %s != (stored %s) %s ' % (type(calculated),
               calculated, type(stored), stored) +\
            '(within %s places), formula: %s' % (self.places, formula)

    def _test_calculator(self):
        self.dframe = self.dataset.dframe()
        row = self.dframe.irow(0)

        columns = self.dframe.columns.tolist()
        self.start_num_cols = len(columns)
        self.added_num_cols = 0

        column_labels_to_slugs = {
            column_attrs[Dataset.LABEL]: (column_name) for
            (column_name, column_attrs) in self.dataset.schema.items()
        }
        self.label_list, self.slugified_key_list = [
            list(ary) for ary in zip(*column_labels_to_slugs.items())
        ]

        for idx, formula in enumerate(self.calculations):
            name = 'test-%s' % idx
            self.parser.validate_formula(formula, row)

            calculator = Calculator(self.dataset)

            groups = self.dataset.split_groups(self.group)
            calculation = Calculation()
            calculation.save(self.dataset, formula, name, self.group)
            calculator.calculate_columns([calculation])

            self.column_labels_to_slugs = self.dataset.schema.labels_to_slugs

            self._test_calculation_results(name, formula)



from bamboo.core.parser import ParseError, Parser
from bamboo.tests.test_base import TestBase


class TestParser(TestBase):

    def setUp(self):
        self.parser = Parser()
        self.row = {'VAR': 1}
        TestBase.setUp(self)

    def _check_func(self, parse_result):
        functions = parse_result
        for func in functions:
            self.assertEqual(func.func.func_name, 'eval')
        return functions[0]

    def test_parse_formula(self):
        func = self._check_func(
            self.parser.parse_formula('VAR'))
        self.assertEqual(func(self.row, self.parser.context), 1)

    def test_bnf(self):
        result = self.parser._Parser__build_bnf()
        self.assertNotEqual(self.parser.bnf, None)

    def test_parse_formula_with_var(self):
        func = self._check_func(
            self.parser.parse_formula('VAR + 1'))
        self.assertEqual(func(self.row, self.parser.context), 2)

    def test_parse_formula_bad_formula(self):
        bad_formulas = [
            '=BAD +++ FOR',
            '2 +>+ 1',
            '1 ** 2',
        ]

        for bad_formula in bad_formulas:
            self.assertRaises(ParseError, self.parser.parse_formula,
                              bad_formula)



from pandas import Series
import simplejson as json

from bamboo.controllers.datasets import Datasets
from bamboo.models.dataset import Dataset
from bamboo.tests.controllers.test_abstract_datasets import\
    TestAbstractDatasets


class TestDatasetsEdits(TestAbstractDatasets):

    def setUp(self):
        TestAbstractDatasets.setUp(self)

    def test_show_row(self):
        dataset_id = self._post_file()
        result = json.loads(self.controller.row_show(dataset_id, 0))

        self.assertTrue(isinstance(result, dict))
        self.assertEqual(9.0, result['amount'])

        result = json.loads(self.controller.row_show(dataset_id, "0"))

        self.assertTrue(isinstance(result, dict))
        self.assertEqual(9.0, result['amount'])

    def test_show_row_nonexistent_index(self):
        dataset_id = self._post_file()
        result = json.loads(self.controller.row_show(dataset_id, "90"))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Datasets.ERROR in result)

    def test_show_row_bad_index(self):
        dataset_id = self._post_file()
        result = json.loads(self.controller.row_show(dataset_id, "A"))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Datasets.ERROR in result)

    def test_delete_row(self):
        dataset_id = self._post_file()
        index = 0
        expected_dframe = Dataset.find_one(
            dataset_id).dframe()[index + 1:].reset_index()
        del expected_dframe['index']

        results = json.loads(self.controller.row_delete(dataset_id, index))
        self.assertTrue(Datasets.SUCCESS in results.keys())

        dataset = Dataset.find_one(dataset_id)
        dframe = dataset.dframe()
        self.assertEqual(self.NUM_ROWS - 1, len(dframe))
        self._check_dframes_are_equal(expected_dframe, dframe)

        # check info updated
        info = dataset.info()
        self.assertEqual(self.NUM_ROWS - 1, info[Dataset.NUM_ROWS])

    def test_update_row(self):
        dataset_id = self._post_file()
        index = 0
        update = {'amount': 10, 'food_type': 'breakfast'}
        expected_dframe = Dataset.find_one(dataset_id).dframe()
        expected_row = expected_dframe.ix[0].to_dict()
        expected_row.update(update)
        expected_dframe.ix[0] = Series(expected_row)

        results = json.loads(self.controller.row_update(dataset_id, index,
                                                        json.dumps(update)))
        self.assertTrue(Datasets.SUCCESS in results.keys())

        dframe = Dataset.find_one(dataset_id).dframe()
        self.assertEqual(self.NUM_ROWS, len(dframe))
        self._check_dframes_are_equal(expected_dframe, dframe)



import json

from bamboo.controllers.version import Version
from bamboo.tests.test_base import TestBase


class TestVersion(TestBase):

    def setUp(self):
        TestBase.setUp(self)
        self.controller = Version()

    def test_index(self):
        response = json.loads(self.controller.index())
        response_keys = response.keys()
        keys = [
            'version',
            'description',
            'branch',
            'commit',
        ]
        for key in keys:
            self.assertTrue(key in response_keys)



import numpy as np

from bamboo.lib.jsontools import series_to_jsondict
from bamboo.lib.mongo import dict_from_mongo, dict_for_mongo


MAX_CARDINALITY_FOR_COUNT = 10000
SUMMARY = 'summary'


class ColumnTypeError(Exception):
    """Exception when grouping on a non-dimensional column."""
    pass


def summarize_series(dtype, data):
    """Call summary function dependent on dtype type.

    :param dtype: The dtype of the column to be summarized.
    :param data: The data to be summarized.

    :returns: The appropriate summarization for the type of `dtype`.
    """
    return {
        np.object_: data.value_counts(),
        np.bool_: data.value_counts(),
        np.float64: data.describe(),
        np.int64: data.describe(),
    }.get(dtype.type, None)


def summarizable(dframe, col, groups, dataset):
    """Check if column should be summarized.

    :param dframe: DataFrame to check unique values in.
    :param col: Column to check for factor and number of uniques.
    :param groups: List of groups if summarizing with group, can be empty.
    :param dataset: Dataset to pull schema from.

    :returns: True if column, with parameters should be summarized, otherwise
        False.
    """
    if dataset.is_factor(col):
        cardinality = dframe[col].nunique() if len(groups) else\
            dataset.cardinality(col)
        if cardinality > MAX_CARDINALITY_FOR_COUNT:
            return False

    return not col in groups


def summarize_df(dframe, groups=[], dataset=None):
    """Calculate summary statistics."""
    dtypes = dframe.dtypes

    return {
        col: {
            SUMMARY: series_to_jsondict(summarize_series(dtypes[col], data))
        } for col, data in dframe.iteritems() if summarizable(
            dframe, col, groups, dataset)
    }


def summarize_with_groups(dframe, groups, dataset):
    """Calculate summary statistics for group."""
    return series_to_jsondict(
        dframe.groupby(groups).apply(summarize_df, groups, dataset))


def summarize(dataset, dframe, groups, no_cache):
    """Raises a ColumnTypeError if grouping on a non-dimensional column."""
    # do not allow group by numeric types
    for group in groups:
        if not dataset.schema.is_dimension(group):
            raise ColumnTypeError("group: '%s' is not a dimension." % group)

    group_str = dataset.join_groups(groups) or dataset.ALL

    # check cached stats for group and update as necessary
    stats = dataset.stats
    group_stats = stats.get(group_str)

    if no_cache or not group_stats:
        group_stats = summarize_with_groups(dframe, groups, dataset) if\
            groups else summarize_df(dframe, dataset=dataset)

        if not no_cache:
            stats.update({group_str: group_stats})
            dataset.update({dataset.STATS: dict_for_mongo(stats)})

    stats_dict = dict_from_mongo(group_stats)

    if groups:
        stats_dict = {group_str: stats_dict}

    return stats_dict



from os import getenv

from pymongo import MongoClient

from bamboo.config import settings


class Database(object):
    """Container for the MongoDB client."""

    # MongoDB client default host and port
    __client__ = MongoClient('localhost', 27017, w=1, j=True)
    __db__ = None

    @classmethod
    def create_db(cls, db_name):
        """Create a database `db_name`.

        :param cls: The class to set a client on.
        :param db_name: The name of the collection to create.
        """
        cls.__db__ = cls.__client__[db_name]

    @classmethod
    def client(cls):
        """Return the database client."""
        return cls.__client__

    @classmethod
    def db(cls, name=None):
        """Create the database if it has not been created.

        If name is provided, create a new database.

        :param cls: The class to set the database for.
        :param name: The name of the database to create, default None.
        :returns: The created database.
        """
        if not cls.__db__ or name:
            cls.create_db(name or settings.DATABASE_NAME)
        return cls.__db__



from collections import defaultdict

from celery.task import task
from pandas import concat

from bamboo.core.aggregator import Aggregator
from bamboo.core.frame import BambooFrame, NonUniqueJoinError
from bamboo.core.parser import ParseError, Parser
from bamboo.lib.mongo import MONGO_RESERVED_KEYS
from bamboo.lib.query_args import QueryArgs
from bamboo.lib.schema_builder import make_unique


class Calculator(object):
    """Perform and store calculations and recalculations on update."""

    dframe = None

    def __init__(self, dataset):
        self.dataset = dataset.reload()
        self.parser = Parser(dataset)

    def validate(self, formula, groups):
        """Validate `formula` and `groups` for calculator's dataset.

        Validate the formula and group string by attempting to get a row from
        the dframe for the dataset and then running parser validation on this
        row. Additionally, ensure that the groups in the group string are
        columns in the dataset.

        :param formula: The formula to validate.
        :param groups: A list of columns to group by.

        :returns: The aggregation (or None) for the formula.
        """
        dframe = self.dataset.dframe(QueryArgs(limit=1))
        row = dframe.irow(0) if len(dframe) else {}

        aggregation = self.parser.validate_formula(formula, row)

        for group in groups:
            if not group in dframe.columns:
                raise ParseError(
                    'Group %s not in dataset columns.' % group)

        return aggregation

    def calculate_columns(self, calculations):
        """Calculate and store new columns for `calculations`.

        The new columns are join t othe Calculation dframe and replace the
        dataset's observations.

        .. note::

            This can result in race-conditions when:

            - deleting ``controllers.Datasets.DELETE``
            - updating ``controllers.Datasets.POST([dataset_id])``

            Therefore, perform these actions asychronously.

        :param calculations: A list of calculations.
        """
        self.__ensure_dframe()
        new_dframe = self.dframe

        for c in calculations:

            if c.aggregation:
                aggregation = self.parse_aggregation(
                    c.formula, c.name, c.groups_as_list)
                aggregation.save()
            else:
                columns = self.parse_columns(c.formula, c.name)
                new_dframe = new_dframe.join(columns[0])

        if len(new_dframe.columns) > len(self.dframe.columns):
            self.dataset.replace_observations(new_dframe)

        # propagate calculation to any merged child datasets
        for merged_dataset in self.dataset.merged_datasets:
            merged_calculator = Calculator(merged_dataset)
            merged_calculator.propagate_column(self.dataset)

    def dependent_columns(self):
        return self.parser.context.dependent_columns

    def propagate_column(self, parent_dataset):
        """Propagate columns in `parent_dataset` to this dataset.

        This is used when there has been a new calculation added to
        a dataset and that new column needs to be propagated to all
        child (merged) datasets.

        :param parent_dataset: The dataset to propagate to `self.dataset`.
        """
        # delete the rows in this dataset from the parent
        self.dataset.remove_parent_observations(parent_dataset.dataset_id)

        # get this dataset without the out-of-date parent rows
        dframe = self.dataset.dframe(keep_parent_ids=True)

        # create new dframe from the upated parent and add parent id
        parent_dframe = parent_dataset.dframe().add_parent_column(
            parent_dataset.dataset_id)

        # merge this new dframe with the existing dframe
        updated_dframe = concat([dframe, parent_dframe])

        # save new dframe (updates schema)
        self.dataset.replace_observations(updated_dframe)
        self.dataset.clear_summary_stats()

        # recur
        for merged_dataset in self.dataset.merged_datasets:
            merged_calculator = Calculator(merged_dataset)
            merged_calculator.propagate_column(self.dataset)

    @task(default_retry_delay=5, ignore_result=True)
    def calculate_updates(self, new_data, new_dframe_raw=None,
                          parent_dataset_id=None, update_id=None):
        """Update dataset with `new_data`.

        This can result in race-conditions when:

        - deleting ``controllers.Datasets.DELETE``
        - updating ``controllers.Datasets.POST([dataset_id])``

        Therefore, perform these actions asychronously.

        :param new_data: Data to update this dataset with.
        :param new_dframe_raw: DataFrame to update this dataset with.
        :param parent_dataset_id: If passed add ID as parent ID to column,
            default is None.
        """
        self.__ensure_dframe()
        self.__ensure_ready(update_id)

        labels_to_slugs = self.dataset.schema.labels_to_slugs

        if new_dframe_raw is None:
            new_dframe_raw = self.dframe_from_update(new_data, labels_to_slugs)

        self._check_update_is_valid(new_dframe_raw)

        new_dframe = new_dframe_raw.recognize_dates_from_schema(
            self.dataset.schema)

        new_dframe, aggregations = self.__add_calcs_and_find_aggregations(
            new_dframe, labels_to_slugs)

        # set parent id if provided
        if parent_dataset_id:
            new_dframe = new_dframe.add_parent_column(parent_dataset_id)

        existing_dframe = self.dataset.dframe(keep_parent_ids=True)

        # merge the two dframes
        updated_dframe = concat([existing_dframe, new_dframe])

        # update (overwrite) the dataset with the new merged dframe
        self.dataset.replace_observations(
            updated_dframe, set_num_columns=False)
        self.dframe = self.dataset.dframe()
        self.dataset.clear_summary_stats()

        self.__update_aggregate_datasets(aggregations, new_dframe)
        self.__update_merged_datasets(new_data, labels_to_slugs)
        self.__update_joined_datasets(new_dframe_raw)

        self.dataset.update_complete(update_id)

    def parse_aggregation(self, formula, name, groups, dframe=None):
        columns = self.parse_columns(formula, name, dframe)

        return Aggregator(self.dataset, self.dframe, groups,
                          self.parser.aggregation, name, columns)

    def parse_columns(self, formula, name, dframe=None):
        """Parse formula into function and variables."""
        if dframe is None:
            dframe = self.dataset.dframe()

        functions = self.parser.parse_formula(formula)

        columns = []

        for function in functions:
            column = dframe.apply(
                function, axis=1, args=(self.parser.context, ))
            column.name = make_unique(name, [c.name for c in columns])
            columns.append(column)

        return columns

    def _check_update_is_valid(self, new_dframe_raw):
        """Check if the update is valid.

        Check whether this is a right-hand side of any joins
        and deny the update if the update would produce an invalid
        join as a result.

        :raises: `NonUniqueJoinError` if update is illegal given joins of
            dataset.
        """
        if any([direction == 'left' for direction, _, on, __ in
                self.dataset.joined_datasets]):
            if on in new_dframe_raw.columns and on in self.dframe.columns:
                merged_join_column = concat(
                    [new_dframe_raw[on], self.dframe[on]])
                if len(merged_join_column) != merged_join_column.nunique():
                    raise NonUniqueJoinError(
                        'Cannot update. This is the right hand join and the'
                        'column "%s" will become non-unique.' % on)

    def __ensure_dframe(self):
        """Ensure `dframe` for the calculator's dataset is defined."""
        if self.dframe is None:
            self.dframe = self.dataset.dframe()

    def __ensure_ready(self, update_id):
        # dataset must not be pending
        if not self.dataset.is_ready or (
                update_id and self.dataset.has_pending_updates(update_id)):
            self.dataset.reload()
            raise self.calculate_updates.retry()

    def __add_calcs_and_find_aggregations(self, new_dframe, labels_to_slugs):
        aggregations = []
        calculations = self.dataset.calculations()

        for calculation in calculations:
            if calculation.aggregation is not None:
                aggregations.append(calculation)
            else:
                function = self.parser.parse_formula(calculation.formula)
                new_column = new_dframe.apply(function[0], axis=1,
                                              args=(self.parser.context, ))
                potential_name = calculation.name

                if potential_name not in self.dframe.columns:
                    if potential_name in labels_to_slugs:
                        new_column.name = labels_to_slugs[potential_name]
                else:
                    new_column.name = potential_name

                new_dframe = new_dframe.join(new_column)

        return new_dframe, aggregations

    def __update_merged_datasets(self, new_data, labels_to_slugs):
        # store slugs as labels for child datasets
        slugified_data = self.__slugify_data(new_data, labels_to_slugs)

        # update the merged datasets with new_dframe
        for mapping, merged_dataset in self.dataset.merged_datasets_with_map:
            merged_calculator = Calculator(merged_dataset)

            slugified_data = self.__remapped_data(mapping, slugified_data)
            merged_calculator.calculate_updates(
                merged_calculator,
                slugified_data,
                parent_dataset_id=self.dataset.dataset_id)

    def __update_joined_datasets(self, new_dframe_raw):
        # update any joined datasets
        for direction, other_dataset, on, joined_dataset in\
                self.dataset.joined_datasets:
            if direction == 'left':
                if on in new_dframe_raw.columns:
                    # only proceed if on in new dframe
                    other_dframe = other_dataset.dframe(padded=True)

                    if len(set(new_dframe_raw[on]).intersection(
                            set(other_dframe[on]))):
                        # only proceed if new on value is in on column in lhs
                        merged_dframe = other_dframe.join_dataset(
                            self.dataset, on)
                        joined_dataset.replace_observations(merged_dframe)
            else:
                merged_dframe = new_dframe_raw

                if on in merged_dframe:
                    merged_dframe = new_dframe_raw.join_dataset(
                        other_dataset, on)

                joined_calculator = Calculator(joined_dataset)
                joined_calculator.calculate_updates(
                    joined_calculator, merged_dframe.to_jsondict(),
                    parent_dataset_id=self.dataset.dataset_id)

    def dframe_from_update(self, new_data, labels_to_slugs):
        """Make a single-row dataframe for the additional data to add.

        :param new_data: Data to add to dframe.
        :type new_data: List.
        :param labels_to_slugs: Map of labels to slugs.
        """
        self.__ensure_dframe()

        filtered_data = []
        columns = self.dframe.columns
        num_columns = len(columns)
        num_rows = len(self.dframe)
        dframe_empty = not num_columns

        if dframe_empty:
            columns = self.dataset.schema.keys()

        for row in new_data:
            filtered_row = dict()
            for col, val in row.iteritems():
                # special case for reserved keys (e.g. _id)
                if col in MONGO_RESERVED_KEYS:
                    if (not num_columns or col in columns) and\
                            col not in filtered_row.keys():
                        filtered_row[col] = val
                else:
                    # if col is a label take slug, if it's a slug take col
                    slug = labels_to_slugs.get(
                        col, col if col in labels_to_slugs.values() else None)

                    # if slug is valid or there is an empty dframe
                    if (slug or col in labels_to_slugs.keys()) and (
                            dframe_empty or slug in columns):
                        filtered_row[slug] = self.dataset.schema.convert_type(
                            slug, val)

            filtered_data.append(filtered_row)

        index = range(num_rows, num_rows + len(filtered_data))

        return BambooFrame(filtered_data, index=index)

    def __update_aggregate_datasets(self, calculations, new_dframe):
        calcs_to_data = self.__create_calculations_to_groups_and_datasets(
            calculations)

        for formula, slug, groups, dataset in calcs_to_data:
            self.__update_aggregate_dataset(formula, new_dframe, slug, groups,
                                            dataset)

    def __update_aggregate_dataset(self, formula, new_dframe, name, groups,
                                   agg_dataset):
        """Update the aggregated dataset built for `self` with `calculation`.

        Proceed with the following steps:

            - delete the rows in this dataset from the parent
            - recalculate aggregated dataframe from aggregation
            - update aggregated dataset with new dataframe and add parent id
            - recur on all merged datasets descending from the aggregated
              dataset

        :param formula: The formula to execute.
        :param new_dframe: The DataFrame to aggregate on.
        :param name: The name of the aggregation.
        :param groups: A column or columns to group on.
        :type group: String, list of strings, or None.
        :param agg_dataset: The DataSet to store the aggregation in.
        """
        # parse aggregation and build column arguments
        aggregation = self.parse_aggregation(formula, name, groups, new_dframe)

        new_agg_dframe = aggregation.update(agg_dataset, self, formula)

        # jsondict from new dframe
        new_data = new_agg_dframe.to_jsondict()

        for merged_dataset in agg_dataset.merged_datasets:
            # remove rows in child from this merged dataset
            merged_dataset.remove_parent_observations(
                agg_dataset.dataset_id)

            # calculate updates on the child
            merged_calculator = Calculator(merged_dataset)
            merged_calculator.calculate_updates(
                merged_calculator, new_data,
                parent_dataset_id=agg_dataset.dataset_id)

    def __create_calculations_to_groups_and_datasets(self, calculations):
        """Create list of groups and calculations."""
        calcs_to_data = defaultdict(list)

        names_to_formulas = {
            calc.name: calc.formula for calc in calculations
        }
        calculations = set([calc.name for calc in calculations])

        for group, dataset in self.dataset.aggregated_datasets:
            labels_to_slugs = dataset.schema.labels_to_slugs
            calculations_for_dataset = list(set(
                labels_to_slugs.keys()).intersection(calculations))

            for calc in calculations_for_dataset:
                calcs_to_data[calc].append((
                    names_to_formulas[calc],
                    labels_to_slugs[calc],
                    group,
                    dataset
                ))

        return [
            item for sublist in calcs_to_data.values() for item in sublist
        ]

    def __slugify_data(self, new_data, labels_to_slugs):
        slugified_data = []

        if not isinstance(new_data, list):
            new_data = [new_data]

        for row in new_data:
            for key, value in row.iteritems():
                if labels_to_slugs.get(key) and key not in MONGO_RESERVED_KEYS:
                    del row[key]
                    row[labels_to_slugs[key]] = value

            slugified_data.append(row)

        return slugified_data

    def __remapped_data(self, mapping, slugified_data):
        column_map = mapping.get(self.dataset.dataset_id) if mapping else None

        if column_map:
            slugified_data = [{column_map.get(k, k): v for k, v in row.items()}
                              for row in slugified_data]

        return slugified_data

    def __getstate__(self):
        """Get state for pickle."""
        return [self.dataset, self.parser]

    def __setstate__(self, state):
        self.dataset, self.parser = state



from nose.tools import assert_raises

from bamboo.core.parser import ParseError
from bamboo.models.calculation import Calculation, DependencyError
from bamboo.models.dataset import Dataset
from bamboo.tests.test_base import TestBase


class TestCalculation(TestBase):

    def setUp(self):
        TestBase.setUp(self)
        self.dataset = Dataset()
        self.dataset.save(self.test_dataset_ids['good_eats.csv'])
        self.formula = 'rating'
        self.name = 'test'

    def _save_calculation(self, formula):
        if not formula:
            formula = self.formula
        return Calculation.create(self.dataset, formula, self.name)

    def _save_observations(self):
        self.dataset.save_observations(self.get_data('good_eats.csv'))

    def _save_observations_and_calculation(self, formula=None):
        self._save_observations()
        return self._save_calculation(formula)

    def test_save(self):
        calculation = self._save_observations_and_calculation()

        self.assertTrue(isinstance(calculation, Calculation))

        record = calculation.record

        self.assertTrue(isinstance(record, dict))
        self.assertTrue(Calculation.FORMULA in record.keys())
        self.assertTrue(Calculation.STATE in record.keys())

        record = Calculation.find(self.dataset)[0].record

        self.assertEqual(record[Calculation.STATE], Calculation.STATE_READY)
        self.assertTrue(Calculation(record).is_ready)

    def test_save_set_status(self):
        record = self._save_observations_and_calculation().record

        self.assertTrue(isinstance(record, dict))
        self.assertTrue(Calculation.FORMULA in record.keys())

    def test_save_set_aggregation(self):
        calculation = self._save_observations_and_calculation('max(amount)')

        self.assertEqual('max', calculation.aggregation)

    def test_save_set_aggregation_id(self):
        calculation = self._save_observations_and_calculation('max(amount)')
        agg_id = self.dataset.aggregated_datasets_dict['']

        self.assertEqual(agg_id, calculation.aggregation_id)

    def test_save_improper_formula(self):
        assert_raises(ParseError, self._save_observations_and_calculation,
                      'NON_EXISTENT_COLUMN')
        try:
            self._save_observations_and_calculation('NON_EXISTENT_COLUMN')
        except ParseError as e:
            self.assertTrue('Missing column' in e.__str__())

    def test_save_unparsable_formula(self):
        assert_raises(ParseError, self._save_observations_and_calculation,
                      '=NON_EXISTENT_COLUMN')
        try:
            self._save_observations_and_calculation(
                '=NON_EXISTENT_COLUMN')
        except ParseError as e:
            self.assertTrue('Parse Failure' in e.__str__())

    def test_save_improper_formula_no_data(self):
        assert_raises(ParseError, Calculation().save, self.dataset,
                      'NON_EXISTENT_COLUMN', self.name)
        try:
            Calculation().save(self.dataset, 'NON_EXISTENT_COLUMN',
                               self.name)
        except ParseError as e:
            self.assertTrue('No schema' in e.__str__())

    def test_save_unparsable_formula_no_data(self):
        assert_raises(ParseError, Calculation().save, self.dataset,
                      '=NON_EXISTENT_COLUMN', self.name)
        try:
            Calculation().save(self.dataset, '=NON_EXISTENT_COLUMN',
                               self.name)
        except ParseError as e:
            self.assertTrue('Parse Failure' in e.__str__())

    def test_save_non_existent_group(self):
        self._save_observations()
        assert_raises(ParseError, Calculation().save, self.dataset,
                      self.formula, self.name, group_str='NON_EXISTENT_GROUP')
        try:
            Calculation().save(self.dataset, self.formula, self.name,
                               group_str='NON_EXISTENT_GROUP')
        except ParseError as e:
            self.assertTrue('Group' in e.__str__())

    def test_find(self):
        record = self._save_observations_and_calculation()
        rows = Calculation.find(self.dataset)
        new_record = rows[0].record
        status = new_record.pop(Calculation.STATE)
        self.assertEqual(status, Calculation.STATE_READY)

    def test_sets_dependent_calculations(self):
        record = self._save_observations_and_calculation()
        self.name = 'test1'
        record = self._save_calculation('test')
        calculation = Calculation.find_one(self.dataset.dataset_id, 'test')
        self.assertEqual(calculation.dependent_calculations, ['test1'])

    def test_removes_dependent_calculations(self):
        record = self._save_observations_and_calculation()
        self.name = 'test1'
        record = self._save_calculation('test')
        calculation = Calculation.find_one(self.dataset.dataset_id, 'test')
        self.assertEqual(calculation.dependent_calculations, ['test1'])
        calculation = Calculation.find_one(self.dataset.dataset_id, 'test1')
        calculation.delete(self.dataset)
        calculation = Calculation.find_one(self.dataset.dataset_id, 'test')
        self.assertEqual(calculation.dependent_calculations, [])

    def test_disallow_delete_dependent_calculation(self):
        record = self._save_observations_and_calculation()
        self.name = 'test1'
        record = self._save_calculation('test')
        calculation = Calculation.find_one(self.dataset.dataset_id, 'test')
        self.assertEqual(calculation.dependent_calculations, ['test1'])
        calculation = Calculation.find_one(self.dataset.dataset_id, 'test')
        assert_raises(DependencyError, calculation.delete, self.dataset)



from datetime import datetime
import numpy as np
import re

from bamboo.core.frame import BAMBOO_RESERVED_KEYS
from bamboo.core.parser import Parser
from bamboo.lib.exceptions import ArgumentError
from bamboo.lib.mongo import MONGO_RESERVED_KEY_STRS, reserve_encoded


CARDINALITY = 'cardinality'
OLAP_TYPE = 'olap_type'
SIMPLETYPE = 'simpletype'
LABEL = 'label'

# olap_types
DIMENSION = 'dimension'
MEASURE = 'measure'

# simpletypes
BOOLEAN = 'boolean'
DATETIME = 'datetime'
INTEGER = 'integer'
FLOAT = 'float'
STRING = 'string'

# map from numpy objects to olap_types
DTYPE_TO_OLAP_TYPE = {
    np.object_: DIMENSION,
    np.bool_: DIMENSION,
    np.float64: MEASURE,
    np.int64: MEASURE,
    datetime: MEASURE,
}

# map from numpy objects to simpletypes
DTYPE_TO_SIMPLETYPE = {
    np.bool_:   BOOLEAN,
    np.float64: FLOAT,
    np.int64:   INTEGER,
    np.object_: STRING,
    datetime: DATETIME,
}
SIMPLETYPE_TO_DTYPE = {
    FLOAT: np.float64,
    INTEGER: np.int64,
}
SIMPLETYPE_TO_OLAP_TYPE = {
    v: DTYPE_TO_OLAP_TYPE[k] for (k, v) in DTYPE_TO_SIMPLETYPE.items()}

RE_ENCODED_COLUMN = re.compile(ur'(?u)\W')


class Schema(dict):
    @classmethod
    def safe_init(cls, arg):
        """Make schema with potential arg of None."""
        return cls() if arg is None else cls(arg)

    @property
    def labels_to_slugs(self):
        """Build dict from column labels to slugs."""
        return {
            column_attrs[LABEL]: reserve_encoded(column_name) for
            (column_name, column_attrs) in self.items()
        }

    @property
    def numeric_slugs(self):
        return [slug for slug, col_schema in self.items()
                if col_schema[SIMPLETYPE] in [INTEGER, FLOAT]]

    def cardinality(self, column):
        if self.is_dimension(column):
            return self[column].get(CARDINALITY)

    def convert_type(self, slug, value):
        column_schema = self.get(slug)
        if column_schema:
            type_func = SIMPLETYPE_TO_DTYPE.get(column_schema[SIMPLETYPE])
            if type_func:
                value = type_func(value)

        return value

    def is_date_simpletype(self, column):
        return self[column][SIMPLETYPE] == DATETIME

    def is_dimension(self, column):
        col_schema = self.get(column)

        return col_schema and col_schema[OLAP_TYPE] == DIMENSION

    def is_measurable(self, column):
        col_schema = self.get(column)

        return col_schema and (
            SIMPLETYPE_TO_OLAP_TYPE[col_schema[SIMPLETYPE]] == MEASURE)

    def rebuild(self, dframe, overwrite=False):
        """Rebuild a schema for a dframe.

        :param dframe: The DataFrame whose schema to merge with the current
            schema.
        :param overwrite: If true replace schema, otherwise update.
        """
        current_schema = self
        new_schema = schema_from_dframe(dframe, self)

        if current_schema and not overwrite:
            # merge new schema with existing schema
            current_schema.update(new_schema)
            new_schema = current_schema

        return new_schema

    def rename_map_for_dframe(self, dframe):
        """Return a map from dframe columns to slugs.

        :param dframe: The DataFrame to produce the map for.
        """
        labels_to_slugs = self.labels_to_slugs

        return {
            column: labels_to_slugs[column] for column in
            dframe.columns.tolist() if self._resluggable_column(
                column, labels_to_slugs, dframe)
        }

    def set_olap_type(self, column, olap_type):
        """Set the OLAP Type for this `column` of schema.

        Only columns with an original OLAP Type of 'measure' can be modified.
        This includes columns with Simple Type integer, float, and datetime.

        :param column: The column to set the OLAP Type for.
        :param olap_type: The OLAP Type to set. Must be 'dimension' or
            'measure'.
        :raises: `ArgumentError` if trying to set the OLAP Type of an column
          whose OLAP Type was not originally a 'measure'.
        """
        if not self.is_measurable(column):
            raise ArgumentError(
                'To set the OLAP Type, column "%s" must be measurable.' %
                column)

        self[column][OLAP_TYPE] = olap_type

    def _resluggable_column(self, column, labels_to_slugs, dframe):
        """Test if column should be slugged.

        A column should be slugged if:
            1. The `column` is a key in `labels_to_slugs` and
            2. The `column` is not a value in `labels_to_slugs` or
                1. The `column` label is not equal to the `column` slug and
                2. The slug is not in the `dframe`'s columns

        :param column: The column to reslug.
        :param labels_to_slugs: The labels to slugs map (only build once).
        :param dframe: The DataFrame that column is in.
        """
        return (column in labels_to_slugs.keys() and (
                not column in labels_to_slugs.values() or (
                    labels_to_slugs[column] != column and
                    labels_to_slugs[column] not in dframe.columns)))


def schema_from_dframe(dframe, schema=None):
    """Build schema from the DataFrame and a schema.

    :param dframe: The DataFrame to build a schema for.
    :param schema: Existing schema, optional.

    :returns: A dictionary schema.
    """
    dtypes = dframe.dtypes.to_dict()

    column_names = list()
    names_to_labels = dict()
    reserved_keys = MONGO_RESERVED_KEY_STRS + BAMBOO_RESERVED_KEYS

    # use existing labels for existing columns
    for name in dtypes.keys():
        if name not in reserved_keys:
            column_names.append(name)
            if schema:
                schema_for_name = schema.get(name)
                if schema_for_name:
                    names_to_labels[name] = schema_for_name[
                        LABEL]

    encoded_names = dict(zip(column_names, _slugify_columns(column_names)))
    schema = Schema()

    for (name, dtype) in dtypes.items():
        if name not in BAMBOO_RESERVED_KEYS:
            column_schema = {
                LABEL: names_to_labels.get(name, name),
                OLAP_TYPE: _olap_type_for_data_and_dtype(
                    dframe[name], dtype),
                SIMPLETYPE: _simpletype_for_data_and_dtype(
                    dframe[name], dtype),
            }
            try:
                column_schema[CARDINALITY] = dframe[
                    name].nunique()
            except AttributeError as e:
                pass
            except TypeError as e:
                # E.g. dates with and without offset can not be compared and
                # raise a type error.
                pass
            schema[encoded_names[name]] = column_schema

    return schema


def _slugify_columns(column_names):
    """Convert list of strings into unique slugs.

    Convert non-alphanumeric characters in column names into underscores and
    ensure that all column names are unique.

    :param column_names: A list of strings.

    :returns: A list of slugified names with a one-to-one mapping to
        `column_names`.
    """

    encoded_names = []

    for column_name in column_names:
        slug = RE_ENCODED_COLUMN.sub('_', column_name).lower()
        slug = make_unique(slug, encoded_names + Parser.reserved_words)
        encoded_names.append(slug)

    return encoded_names


def make_unique(name, reserved_names):
    """Return a slug ensuring name is not in `reserved_names`.

    :param name: The name to make unique.
    :param reserved_names: A list of names the column must not be included in.
    """
    while name in reserved_names:
        name += '_'

    return name


def _olap_type_for_data_and_dtype(column, dtype):
    return _type_for_data_and_dtypes(
        DTYPE_TO_OLAP_TYPE, column, dtype.type)


def _simpletype_for_data_and_dtype(column, dtype):
    return _type_for_data_and_dtypes(
        DTYPE_TO_SIMPLETYPE, column, dtype.type)


def _type_for_data_and_dtypes(type_map, column, dtype_type):
    has_datetime = any([isinstance(field, datetime) for field in column])

    return type_map[datetime if has_datetime else dtype_type]



import os

ASYNC_FLAG = 'BAMBOO_ASYNC_OFF'


def is_async():
    return not os.getenv(ASYNC_FLAG)


def set_async(on):
    if on:
        if not is_async():
            del os.environ[ASYNC_FLAG]
    else:
        os.environ[ASYNC_FLAG] = 'True'


def call_async(function, *args, **kwargs):
    """Potentially asynchronously call `function` with the arguments.

    :param function: The function to call.
    :param args: Arguments for the function.
    :param kwargs: Keyword arguments for the function.
    """
    countdown = kwargs.pop('countdown', 0)

    if is_async():
        function.__getattribute__('apply_async')(
            countdown=countdown, args=args, kwargs=kwargs)
    else:  # pragma: no cover
        function(*args, **kwargs)



from bamboo.controllers.abstract_controller import AbstractController
from bamboo.controllers.calculations import Calculations
from bamboo.controllers.datasets import Datasets
from bamboo.tests.test_base import TestBase


class TestOptions(TestBase):
    # TODO: check response code == 204
    # and headers are set correctly...

    def setUp(self):
        TestBase.setUp(self)
        self.controller = AbstractController()

    def _test_options(self, response):
        self.assertTrue(response == '')

    def test_abstract_controller_options(self):
        response = self.controller.options()

    def test_dataset_controller_options(self):
        controller = Datasets()
        response = controller.options('dataset_id')

    def test_calculations_controller_options(self):
        controller = Calculations()
        response = self.controller.options('dataset_id')



from datetime import datetime

from pandas import Series

from bamboo.core.frame import BAMBOO_RESERVED_KEYS, BambooFrame,\
    PARENT_DATASET_ID
from bamboo.lib.schema_builder import DATETIME, SIMPLETYPE, Schema
from bamboo.lib.mongo import MONGO_RESERVED_KEYS
from bamboo.tests.test_base import TestBase


class TestFrame(TestBase):

    def setUp(self):
        TestBase.setUp(self)
        self.dframe = self.get_data('good_eats.csv')
        self.bframe = BambooFrame(self.dframe)

    def _add_bamboo_reserved_keys(self, value=1):
        for key in BAMBOO_RESERVED_KEYS:
            column = Series([value] * len(self.bframe))
            column.name = key
            self.bframe = BambooFrame(self.bframe.join(column))

    def test_add_parent_column(self):
        value = 1
        self._add_bamboo_reserved_keys(value)
        for index, item in self.bframe[PARENT_DATASET_ID].iteritems():
            self.assertEqual(item, value)

    def test_decode_mongo_reserved_keys(self):
        prev_columns = self.bframe.columns
        for col in MONGO_RESERVED_KEYS:
            self.assertTrue(col in self.bframe.columns)
        self.bframe.decode_mongo_reserved_keys()
        for col in MONGO_RESERVED_KEYS:
            self.assertFalse(col in self.bframe.columns)

    def test_recognize_dates(self):
        bframe_with_dates = self.bframe.recognize_dates()
        for field in bframe_with_dates['submit_date']:
            self.assertTrue(isinstance(field, datetime))

    def test_recognize_dates_from_schema(self):
        schema = Schema({
            'submit_date': {
                SIMPLETYPE: DATETIME
            }
        })
        bframe_with_dates = self.bframe.recognize_dates_from_schema(schema)
        for field in bframe_with_dates['submit_date']:
            self.assertTrue(isinstance(field, datetime))

    def test_remove_bamboo_reserved_keys(self):
        self._add_bamboo_reserved_keys()
        for key in BAMBOO_RESERVED_KEYS:
            self.assertTrue(key in self.bframe.columns)
        self.bframe.remove_bamboo_reserved_keys()
        for key in BAMBOO_RESERVED_KEYS:
            self.assertFalse(key in self.bframe.columns)

    def test_remove_bamboo_reserved_keys_exclusion(self):
        self._add_bamboo_reserved_keys()
        for key in BAMBOO_RESERVED_KEYS:
            self.assertTrue(key in self.bframe.columns)
        self.bframe.remove_bamboo_reserved_keys([PARENT_DATASET_ID])
        for key in BAMBOO_RESERVED_KEYS:
            if key == PARENT_DATASET_ID:
                self.assertTrue(key in self.bframe.columns)
            else:
                self.assertFalse(key in self.bframe.columns)

    def test_only_rows_for_parent_id(self):
        parent_id = 1
        len_parent_rows = len(self.bframe) / 2

        column = Series([parent_id] * len_parent_rows)
        column.name = PARENT_DATASET_ID

        self.bframe = BambooFrame(self.bframe.join(column))
        bframe_only = self.bframe.only_rows_for_parent_id(parent_id)

        self.assertFalse(PARENT_DATASET_ID in bframe_only.columns)
        self.assertEqual(len(bframe_only), len_parent_rows)

    def test_to_jsondict(self):
        jsondict = self.bframe.to_jsondict()
        self.assertEqual(len(jsondict), len(self.bframe))
        for col in jsondict:
            self.assertEqual(len(col), len(self.bframe.columns))

    def test_to_json(self):
        json = self.bframe.to_json()
        self.assertEqual(type(json), str)



class QueryArgs(object):
    @property
    def order_by(self):
        order_by = self.order_by_value

        if order_by:
            if order_by[0] in ('-', '+'):
                sort_dir, field = -1 if order_by[0] == '-' else 1, order_by[1:]
            else:
                sort_dir, field = 1, order_by
            order_by = [(field, sort_dir)]

        return order_by

    def __init__(self, query={}, select=None, distinct=None, limit=0,
                 order_by=None):
        """A holder for query arguments.

        :param query: An optional query.
        :param select: An optional select to limit the fields in the dframe.
        :param distinct: Return distinct entries for this field.
        :param limit: Limit on the number of rows in the returned dframe.
        :param order_by: Sort resulting rows according to a column value and
            sign indicating ascending or descending.

        Example of `order_by`:

          - ``order_by='mycolumn'``
          - ``order_by='-mycolumn'``
        """

        self.query = query
        self.select = select
        self.distinct = distinct
        self.limit = limit
        self.order_by_value = order_by



import os
import sys

from bamboo.core.frame import *
from bamboo.lib.async import *


# test settings
TESTING = 'test' in sys.argv[0].split('/')[-1]

os.environ['CELERY_CONFIG_MODULE'] = 'bamboo.config.celeryconfig'

if TESTING:
    set_async(False)
    os.environ['CELERY_CONFIG_MODULE'] += '_test'

# perform import after env variables are set
from bamboo.models.dataset import Dataset



import simplejson as json

from bamboo.controllers.calculations import Calculations
from bamboo.controllers.datasets import Datasets
from bamboo.core.frame import BambooFrame
from bamboo.core.summary import SUMMARY
from bamboo.lib.jsontools import series_to_jsondict
from bamboo.lib.mongo import MONGO_RESERVED_KEYS
from bamboo.models.dataset import Dataset
from bamboo.tests.test_base import TestBase


class TestAbstractDatasets(TestBase):

    NUM_COLS = 15
    NUM_ROWS = 19

    def setUp(self):
        TestBase.setUp(self)
        self.controller = Datasets()
        self._file_name = 'good_eats.csv'
        self._update_file_name = 'good_eats_update.json'
        self._update_check_file_path = '%sgood_eats_update_values.json' % (
            self.FIXTURE_PATH)
        self.default_formulae = [
            'amount',
            'amount + 1',
            'amount - 5',
        ]

    def _put_row_updates(self, dataset_id=None, file_name=None, validate=True):
        if not dataset_id:
            dataset_id = self.dataset_id

        if not file_name:
            file_name = self._update_file_name

        update = open('%s%s' % (self.FIXTURE_PATH, file_name), 'r').read()
        result = json.loads(self.controller.update(dataset_id=dataset_id,
                                                   update=update))

        if validate:
            self.assertTrue(isinstance(result, dict))
            self.assertTrue(Dataset.ID in result)

        # set up the (default) values to test against
        with open(self._update_check_file_path, 'r') as f:
            self._update_values = json.loads(f.read())

    def _load_schema(self):
        return json.loads(
            self.controller.info(self.dataset_id))[Dataset.SCHEMA]

    def _check_dframes_are_equal(self, dframe1, dframe2):
        self._check_dframe_is_subset(dframe1, dframe2)
        self._check_dframe_is_subset(dframe2, dframe1)

    def _check_dframe_is_subset(self, dframe1, dframe2):
        dframe2_rows = [self._reduce_precision(row) for row in
                        BambooFrame(dframe2).to_jsondict()]

        for row in dframe1.iterrows():
            dframe1_row = self._reduce_precision(series_to_jsondict(row[1]))
            self.assertTrue(dframe1_row in dframe2_rows,
                            'dframe1_row: %s\n\ndframe2_rows: %s' % (
                                dframe1_row, dframe2_rows))

    def _reduce_precision(self, row):
        for key, value in row.iteritems():
            if isinstance(value, float):
                row[key] = round(value, 10)

        return row

    def _post_calculations(self, formulae=[], group=None):
        schema = self._load_schema()
        controller = Calculations()

        for idx, formula in enumerate(formulae):
            name = 'calc_%d' % idx if not schema or\
                formula in schema.keys() else formula

            controller.create(self.dataset_id, formula=formula, name=name,
                              group=group)

    def _test_summary_built(self, result):
        # check that summary is created
        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Dataset.ID in result)
        self.dataset_id = result[Dataset.ID]

        results = self.controller.summary(
            self.dataset_id,
            select=self.controller.SELECT_ALL_FOR_SUMMARY)

        return self._test_summary_results(results)

    def _test_summary_results(self, results):
        results = json.loads(results)
        self.assertTrue(isinstance(results, dict))
        return results

    def _test_aggregations(self, groups=['']):
        results = json.loads(self.controller.aggregations(self.dataset_id))
        self.assertTrue(isinstance(results, dict))
        self.assertEqual(len(results.keys()), len(groups))
        self.assertEqual(results.keys(), groups)
        linked_dataset_id = results[groups[0]]
        self.assertTrue(isinstance(linked_dataset_id, basestring))

        # inspect linked dataset
        return json.loads(self.controller.show(linked_dataset_id))

    def _test_summary_no_group(self, results, dataset_id=None, group=None):
        if not dataset_id:
            dataset_id = self.dataset_id

        group = [group] if group else []
        result_keys = results.keys()

        # minus the column that we are grouping on
        self.assertEqual(len(result_keys), self.NUM_COLS - len(group))

        columns = [col for col in
                   self.get_data(self._file_name).columns.tolist()
                   if not col in MONGO_RESERVED_KEYS + group]

        dataset = Dataset.find_one(dataset_id)
        labels_to_slugs = dataset.schema.labels_to_slugs

        for col in columns:
            slug = labels_to_slugs[col]
            self.assertTrue(slug in result_keys,
                            'col (slug): %s in: %s' % (slug, result_keys))
            self.assertTrue(SUMMARY in results[slug].keys())



import simplejson as json

from bamboo.models.dataset import Dataset
from bamboo.tests.controllers.test_abstract_datasets_update import\
    TestAbstractDatasetsUpdate


class TestDatasetsUpdateWithCalcs(TestAbstractDatasetsUpdate):

    def setUp(self):
        TestAbstractDatasetsUpdate.setUp(self)
        self._create_original_datasets()
        self._add_common_calculations()

        # create linked datasets
        self.calculations.create(
            self.dataset2_id, 'sum(amount)', 'sum of amount')
        result = json.loads(
            self.controller.aggregations(self.dataset2_id))
        self.linked_dataset1_id = result['']

        # create merged datasets
        result = json.loads(self.controller.merge(dataset_ids=json.dumps(
            [self.dataset1_id, self.dataset2_id])))
        self.merged_dataset1_id = result[Dataset.ID]

        result = json.loads(self.controller.merge(dataset_ids=json.dumps(
            [self.merged_dataset1_id, self.linked_dataset1_id])))
        self.merged_dataset2_id = result[Dataset.ID]

    def test_setup_datasets(self):
        self._verify_dataset(
            self.dataset1_id,
            'updates_with_calcs/originals/dataset1.pkl')
        self._verify_dataset(
            self.dataset2_id,
            'updates_with_calcs/originals/dataset2.pkl')
        self._verify_dataset(
            self.linked_dataset1_id,
            'updates_with_calcs/originals/linked_dataset1.pkl')
        self._verify_dataset(
            self.merged_dataset1_id,
            'updates_with_calcs/originals/merged_dataset1.pkl')
        self._verify_dataset(
            self.merged_dataset2_id,
            'updates_with_calcs/originals/merged_dataset2.pkl')

    def _add_calculations(self):
        self.calculations.create(self.dataset2_id,
                                 'amount_plus_gps_alt > gps_precision',
                                 'amount plus gps_alt > gps_precision')
        self.calculations.create(self.linked_dataset1_id,
                                 'sum_of_amount * 2',
                                 'amount')
        self.calculations.create(self.merged_dataset1_id,
                                 'gps_alt * 2',
                                 'double gps_alt')
        self.calculations.create(self.merged_dataset2_id,
                                 'amount * 2',
                                 'double amount')
        self._verify_dataset(
            self.dataset2_id,
            'updates_with_calcs/calcs/dataset2.pkl')
        self._verify_dataset(
            self.linked_dataset1_id,
            'updates_with_calcs/calcs/linked_dataset1.pkl')
        self._verify_dataset(
            self.merged_dataset1_id,
            'updates_with_calcs/calcs/merged_dataset1.pkl')
        self._verify_dataset(
            self.merged_dataset2_id,
            'updates_with_calcs/calcs/merged_dataset2.pkl')

    def test_datasets_add_calculations(self):
        self._add_calculations()



from bamboo.controllers.calculations import Calculations
from bamboo.controllers.datasets import Datasets
from bamboo.controllers.root import Root
from bamboo.controllers.version import Version

# define routes as tuples:
# (name, method, route, controller, action)
ROUTES = [
    # Options for Cross Origin Resource Sharing (CORS)
    ('dataset_options', 'OPTIONS',
        '/datasets/:dataset_id', 'datasets', 'options'),
    ('calculations_options', 'OPTIONS',
        '/datasets/:dataset_id/calculations', 'calculations', 'options'),
    ('calculations_options_alias', 'OPTIONS',
        '/datasets/:dataset_id/calculations/:name', 'calculations', 'options'),
    # root
    ('root', 'GET',
        '/', 'root', 'index'),
    # datasets
    ('datasets_aggregations', 'GET',
        '/datasets/:dataset_id/aggregations', 'datasets', 'aggregations'),
    ('datasets_create', 'POST',
        '/datasets', 'datasets', 'create'),
    ('datasets_delete', 'DELETE',
        '/datasets/:dataset_id', 'datasets', 'delete'),
    ('datasets_drop_columns', 'PUT',
        '/datasets/:dataset_id/drop_columns', 'datasets', 'drop_columns'),
    ('datasets_info', 'GET',
        '/datasets/:dataset_id/info', 'datasets', 'info'),
    ('datasets_merge', 'POST',
        '/datasets/merge', 'datasets', 'merge'),
    ('datasets_join', 'POST',
        '/datasets/:dataset_id/join', 'datasets', 'join'),
    ('datasets_join_alias', 'POST',
        '/datasets/join', 'datasets', 'join'),
    ('datasets_resample', 'GET',
        '/datasets/:dataset_id/resample', 'datasets', 'resample'),
    ('datasets_rolling', 'GET',
        '/datasets/:dataset_id/rolling', 'datasets', 'rolling'),
    ('datasets_set_olap_type', 'PUT',
        '/datasets/:dataset_id/set_olap_type', 'datasets', 'set_olap_type'),
    ('datasets_show', 'GET',
        '/datasets/:dataset_id.:format', 'datasets', 'show'),
    ('datasets_show', 'GET',
        '/datasets/:dataset_id', 'datasets', 'show'),
    ('datasets_set_info', 'PUT',
        '/datasets/:dataset_id/info', 'datasets', 'set_info'),
    ('datasets_summary', 'GET',
        '/datasets/:dataset_id/summary', 'datasets', 'summary'),
    ('datasets_update', 'PUT',
        '/datasets/:dataset_id', 'datasets', 'update'),
    ('datasets_row_delete', 'DELETE', '/datasets/:dataset_id/row/:index',
        'datasets', 'row_delete'),
    ('datasets_row_show', 'GET', '/datasets/:dataset_id/row/:index',
        'datasets', 'row_show'),
    ('datasets_row_update', 'PUT', '/datasets/:dataset_id/row/:index',
        'datasets', 'row_update'),
    # calculations
    ('calculations_create', 'POST',
        '/calculations/:dataset_id', 'calculations', 'create'),
    ('calculations_create_alias', 'POST',
        '/datasets/:dataset_id/calculations', 'calculations', 'create'),
    ('calculations_delete', 'DELETE',
        '/datasets/:dataset_id/calculations', 'calculations', 'delete'),
    ('calculations_delete_alias', 'DELETE',
        '/datasets/:dataset_id/calculations/:name', 'calculations', 'delete'),
    ('calculations_show', 'GET',
        '/calculations/:dataset_id', 'calculations', 'show'),
    ('calculations_show_alias', 'GET',
        '/datasets/:dataset_id/calculations', 'calculations', 'show'),
    # version
    ('version', 'GET',
        '/version', 'version', 'index'),
]


def connect_routes(dispatcher):
    """This function takes the dispatcher and attaches the routes.

    :param dispatcher: The CherryPy dispatcher.
    """
    # controller instances map
    controllers = {
        'root': Root(),
        'calculations': Calculations(),
        'datasets': Datasets(),
        'version': Version(),
    }

    # map them into args to dispatcher
    dictify = lambda x: dict(zip(
        ['name', 'conditions', 'route', 'controller', 'action'], x))
    route_case = {
        'conditions': lambda v: dict(method=v),
        'controller': lambda v: controllers[v],
    }
    kwarg_map = lambda d: {
        k: route_case.get(k, lambda v: v)(v) for k, v in d.iteritems()
    }

    routes = [kwarg_map(dictify(route)) for route in ROUTES]

    # attach them
    for route in routes:
        dispatcher.connect(**route)



from subprocess import check_output

# versioning
VERSION_MAJOR = 0.5
VERSION_MINOR = 7
VERSION_NUMBER = '%.1f.%d' % (VERSION_MAJOR, VERSION_MINOR)
VERSION_DESCRIPTION = 'alpha'


def safe_command_request(args):
    try:
        return check_output(args).strip()
    except:
        # might fail at least if git is not present
        # or if there's no git repository
        return ''


def get_version():
    return {'version': VERSION_NUMBER,
            'version_major': VERSION_MAJOR,
            'version_minor': VERSION_MINOR,
            'description': VERSION_DESCRIPTION,
            'branch': safe_command_request([
                'git', 'rev-parse', '--abbrev-ref', 'HEAD']),
            'commit': safe_command_request(['git', 'rev-parse', 'HEAD'])}



import pickle

from bamboo.controllers.calculations import Calculations
from bamboo.lib.datetools import recognize_dates
from bamboo.models.dataset import Dataset
from bamboo.tests.controllers.test_abstract_datasets import\
    TestAbstractDatasets


class TestAbstractDatasetsUpdate(TestAbstractDatasets):

    def setUp(self):
        """
        These tests use the following dataset configuration:

            d -> dataset
            m -> merged
            l -> linked

            d1   d2
             \  /  \
              m1    l1
                \  /
                 m2

        Dependencies flow from top to bottom.
        """
        TestAbstractDatasets.setUp(self)

    def _create_original_datasets(self):
        self.dataset1_id = self._post_file()
        self.dataset2_id = self._post_file()

    def _add_common_calculations(self):
        self.calculations = Calculations()
        self.calculations.create(
            self.dataset2_id, 'amount + gps_alt', 'amount plus gps_alt')

    def _verify_dataset(self, dataset_id, fixture_path):
        dframe = Dataset.find_one(dataset_id).dframe()
        expected_dframe = recognize_dates(
            pickle.load(open('%s%s' % (
                self.FIXTURE_PATH, fixture_path), 'rb')))
        self._check_dframes_are_equal(dframe, expected_dframe)



import imp
from setuptools import setup

imp.load_source('bamboo_version', 'bamboo/lib/version.py')
from bamboo_version import VERSION_NUMBER

setup(
    name='bamboo-server',
    version='0.5.7.8',
    author='Modi Research Group',
    author_email='info@modilabs.org',
    packages=['bamboo',
              'bamboo.config',
              'bamboo.controllers',
              'bamboo.core',
              'bamboo.lib',
              'bamboo.models',
              'bamboo.tests.core',
              'bamboo.tests.controllers',
              'bamboo.tests.lib',
              'bamboo.tests.models',
              ],
    package_data={'bamboo.tests': ['tests/fixtures/*.csv',
                                   'tests/fixtures/*.json',
                                   ],
                  },
    scripts=['scripts/bamboo.sh',
             'scripts/bamboo_uwsgi.sh',
             'scripts/commands.sh',
             'scripts/install.sh',
             'scripts/run_server.py',
             'scripts/test.sh',
             'scripts/timeit.sh',
             'celeryd/celeryd',
             'celeryd/celeryd-centos'],
    url='http://bamboo.io',
    description='Dynamic data analysis over the web. The logic to your data '
                'dashboards.',
    long_description=open('README.rst', 'rt').read(),
    install_requires=[
        # celery requires python-dateutil>=1.5,<2.0
        'python-dateutil==1.5',

        # for pandas
        'numpy',
        'pandas',
        'scipy',

        # for celery
        'kombu',
        'celery',
        'pymongo',

        'cherrypy',
        'pyparsing',
        'simplejson',
        'Routes'
    ],
)



import simplejson as json

from bamboo.controllers.datasets import Datasets
from bamboo.tests.controllers.test_abstract_datasets import\
    TestAbstractDatasets


class TestDatasetsTsFunctions(TestAbstractDatasets):

    def setUp(self):
        TestAbstractDatasets.setUp(self)

    def __build_resample_result(self, query=None):
        dataset_id = self._post_file('good_eats.csv')
        date_column = 'submit_date'
        interval = 'W'
        results = json.loads(self.controller.resample(
            dataset_id, date_column, interval, query=query))

        self.assertTrue(isinstance(results, list))

        return [date_column, results]

    def __check_interval(self, date_column, results):
        last_date_time = None

        for row in results:
            new_date_time = row[date_column]['$date']

            if last_date_time:
                self.assertEqual(604800000, new_date_time - last_date_time)

            last_date_time = new_date_time

    def test_resample_only_shows_numeric(self):
        date_column, results = self.__build_resample_result()

        permitted_keys = [
            '_id',
            '_percentage_complete',
            'gps_alt',
            'gps_precision',
            'amount',
            'gps_latitude',
            'gps_longitude',
        ] + [date_column]

        for result in results:
            for key in result.keys():
                self.assertTrue(key in permitted_keys)

    def test_resample_interval_correct(self):
        date_column, results = self.__build_resample_result()

        self.__check_interval(date_column, results)

    def test_resample_non_date_column(self):
        dataset_id = self._post_file('good_eats.csv')
        result = json.loads(self.controller.resample(
            dataset_id, 'amount', 'W'))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Datasets.ERROR in result)
        self.assertTrue('DatetimeIndex' in result[Datasets.ERROR])

    def test_resample_bad_interval(self):
        dataset_id = self._post_file('good_eats.csv')
        interval = 'BAD'
        result = json.loads(self.controller.resample(
            dataset_id, 'submit_date', interval))

        self.assertTrue(isinstance(result, dict))
        self.assertTrue(Datasets.ERROR in result)
        self.assertEqual(
            'Could not evaluate %s' % interval, result[Datasets.ERROR])

    def test_resample_with_query(self):
        expected_length = 15
        date_column, results = self.__build_resample_result()

        self.assertEqual(expected_length, len(results))

    def test_resample_with_query(self):
        expected_length = 15
        query = '{"food_type": "lunch"}'
        date_column, results = self.__build_resample_result(query)

        self.assertEqual(expected_length, len(results))

        self.__check_interval(date_column, results)

    def test_rolling_mean(self):
        dataset_id = self._post_file('good_eats.csv')
        window = '3'
        results = json.loads(self.controller.rolling(
            dataset_id, window))
        self.assertTrue(isinstance(results, list))

        for i, row in enumerate(results):
            if i < int(window) - 1:
                for value in row.values():
                    self.assertEqual('null', value)
            else:
                self.assertTrue(isinstance(row['amount'], float))

    def test_rolling_bad_window(self):
        dataset_id = self._post_file('good_eats.csv')
        window = '3n'
        results = json.loads(self.controller.rolling(
            dataset_id, window))
        self.assertTrue(Datasets.ERROR in results.keys())

    def test_rolling_bad_type(self):
        dataset_id = self._post_file('good_eats.csv')
        results = json.loads(self.controller.rolling(
            dataset_id, 3, win_type='BAD'))

        self.assertTrue(isinstance(results, dict))
        self.assertTrue(Datasets.ERROR in results)
        self.assertEqual('Unknown window type.', results[Datasets.ERROR])



import numpy as np

from bamboo.models.dataset import Dataset
from bamboo.tests.core.test_calculator import TestCalculator


class TestCalculations(TestCalculator):

    def setUp(self):
        TestCalculator.setUp(self)
        self.calculations = [
            # constants
            '-9 + 5',

            # aliases
            'rating',
            'gps',

            # arithmetic
            'amount + gps_alt',
            'amount - gps_alt',
            'amount + 5',
            'amount - gps_alt + 2.5',
            'amount * gps_alt',
            'amount / gps_alt',
            'amount * gps_alt / 2.5',
            'amount + gps_alt * gps_precision',

            # precedence
            '(amount + gps_alt) * gps_precision',

            # comparison
            'amount == 2',
            '10 < amount',
            '10 < amount + gps_alt',

            # logical
            'not amount == 2',
            'not(amount == 2)',
            'amount == 2 and 10 < amount',
            'amount == 2 or 10 < amount',
            'not not amount == 2 or 10 < amount',
            'not amount == 2 or 10 < amount',
            '(not amount == 2) or 10 < amount',
            'not(amount == 2 or 10 < amount)',
            'amount ^ 3',
            '(amount + gps_alt) ^ 2 + 100',
            '-amount',
            '-amount < gps_alt - 100',

            # membership
            'rating in ["delectible"]',
            'risk_factor in ["low_risk"]',
            'amount in ["9.0", "2.0", "20.0"]',
            '(risk_factor in ["low_risk"]) and (amount in ["9.0", "20.0"])',

            # dates
            'date("09-04-2012") - submit_date > 21078000',

            # cases
            'case food_type in ["morning_food"]: 1, food_type in ["lunch"]: 2,'
            ' default: 3',
            'case food_type in ["morning_food"]: 1, food_type in ["lunch"]: 2',

            # row-wise column-based aggregations
            'percentile(amount)',
        ]

    def _test_calculation_results(self, name, formula):
            unslug_name = name
            name = self.column_labels_to_slugs[unslug_name]

            # test that updated dataframe persisted
            self.dframe = self.dataset.dframe()
            self.assertTrue(name in self.dframe.columns)

            # test new number of columns
            self.added_num_cols += 1
            self.assertEqual(self.start_num_cols + self.added_num_cols,
                             len(self.dframe.columns.tolist()))

            # test that the schema is up to date
            dataset = Dataset.find_one(self.dataset.dataset_id)
            self.assertTrue(Dataset.SCHEMA in dataset.record.keys())
            self.assertTrue(isinstance(dataset.schema, dict))
            schema = dataset.schema

            # test slugified column names
            self.slugified_key_list.append(name)
            self.assertEqual(sorted(schema.keys()),
                             sorted(self.slugified_key_list))

            # test column labels
            self.label_list.append(unslug_name)
            labels = [schema[col][Dataset.LABEL] for col in schema.keys()]
            self.assertEqual(sorted(labels), sorted(self.label_list))

            # test result of calculation
            formula = self.column_labels_to_slugs[formula]

            for idx, row in self.dframe.iterrows():
                try:
                    result = np.float64(row[name])
                    stored = np.float64(row[formula])
                    # np.nan != np.nan, continue if we have two nan values
                    if np.isnan(result) and np.isnan(stored):
                        continue
                    msg = self._equal_msg(result, stored, formula)
                    self.assertAlmostEqual(result, stored, self.places, msg)
                except ValueError:
                    msg = self._equal_msg(row[name], row[formula], formula)
                    self.assertEqual(row[name], row[formula], msg)

    def test_calculator(self):
        self._test_calculator()



from bamboo.lib.datetools import recognize_dates
from bamboo.tests.test_base import TestBase


class TestDatetools(TestBase):

    def setUp(self):
        TestBase.setUp(self)
        self.dframe = self.get_data('soil_samples.csv')

    def test_recognize_dates(self):
        with_dates = recognize_dates(self.dframe)
        for field in with_dates['single_letter']:
            self.assertTrue(isinstance(field, basestring))



from cherrypy import HTTPRedirect

from bamboo.controllers.root import Root
from bamboo.tests.test_base import TestBase


class TestRoot(TestBase):

    def setUp(self):
        TestBase.setUp(self)
        self.controller = Root()

    def test_index(self):
        try:
            response = self.controller.index()
        except HTTPRedirect as redirect:
            self.assertEqual(redirect.status, 303)
            self.assertTrue(redirect.urls[0].endswith('/docs/index.html'))



#!/usr/bin/env python

import sys
sys.stdout = sys.stderr
sys.path.append('/var/www/bamboo/current')
sys.path.append('/var/www/bamboo/current/bamboo')

import cherrypy

from bamboo.config.routes import connect_routes


# use routes dispatcher
dispatcher = cherrypy.dispatch.RoutesDispatcher()
routes_conf = {'/': {'request.dispatch': dispatcher}}
prod_conf = 'config/prod.conf'

# connect routes
connect_routes(dispatcher)

# global config
cherrypy.config.update(routes_conf)
cherrypy.config.update(prod_conf)
cherrypy.config.update({'environment': 'embedded'})

# app config
application = cherrypy.tree.mount(root=None, config=routes_conf)
application.merge(prod_conf)



from celery.task import task
from pandas import concat

from bamboo.core.frame import BambooFrame
from bamboo.lib.io import import_dataset
from bamboo.lib.async import call_async
from bamboo.models.dataset import Dataset


class MergeError(Exception):
    """For errors while merging datasets."""
    pass


def merge_dataset_ids(dataset_ids, mapping):
    """Load a JSON array of dataset IDs and start a background merge task.

    :param dataset_ids: An array of dataset IDs to merge.

    :raises: `MergeError` if less than 2 datasets are provided. If a dataset
        cannot be found for a dataset ID it is ignored. Therefore if 2 dataset
        IDs are provided and one of them is bad an error is raised.  However,
        if three dataset IDs are provided and one of them is bad, an error is
        not raised.
    """
    datasets = [Dataset.find_one(dataset_id) for dataset_id in dataset_ids]
    datasets = [dataset for dataset in datasets if dataset.record]

    if len(datasets) < 2:
        raise MergeError(
            'merge requires 2 datasets (found %s)' % len(datasets))

    new_dataset = Dataset.create()

    call_async(__merge_datasets_task, new_dataset, datasets, mapping)

    return new_dataset


@task(default_retry_delay=2, ignore_result=True)
def __merge_datasets_task(new_dataset, datasets, mapping):
    """Merge datasets specified by dataset_ids.

    :param new_dataset: The dataset store the merged dataset in.
    :param dataset_ids: A list of IDs to merge into `new_dataset`.
    """
    # check that all datasets are in a 'ready' state
    while any([not dataset.record_ready for dataset in datasets]):
        [dataset.reload() for dataset in datasets]
        raise __merge_datasets_task.retry(countdown=1)

    new_dframe = __merge_datasets(datasets, mapping)

    # save the resulting dframe as a new dataset
    new_dataset.save_observations(new_dframe)

    # store the child dataset ID with each parent
    for dataset in datasets:
        dataset.add_merged_dataset(mapping, new_dataset)


def __merge_datasets(datasets, mapping):
    """Merge two or more datasets."""
    dframes = []

    if not mapping:
        mapping = {}

    for dataset in datasets:
        dframe = dataset.dframe()
        column_map = mapping.get(dataset.dataset_id)

        if column_map:
            dframe = BambooFrame(dframe.rename(columns=column_map))

        dframe = dframe.add_parent_column(dataset.dataset_id)
        dframes.append(dframe)

    return concat(dframes, ignore_index=True)



from bamboo.config import settings

BROKER_BACKEND = 'mongodb'
BROKER_URL = 'mongodb://localhost:27017/%s' % settings.DATABASE_NAME
CELERY_RESULT_BACKEND = 'mongodb'
CELERY_MONGODB_BACKEND_SETTINGS = {
    'host': 'localhost',
    'port': 27017,
    'database': settings.DATABASE_NAME,
    'taskmeta_collection': 'celery_tasks',
}
CELERY_IMPORTS = (
    'bamboo.core.merge',
    'bamboo.lib.io',
    'bamboo.models.calculation',
    'bamboo.models.dataset',
)
CELERYD_CONCURRENCY = 1



from collections import defaultdict
import pickle

import numpy as np

from bamboo.tests.core.test_calculator import TestCalculator


class TestAggregations(TestCalculator):

    AGGREGATION_RESULTS = {
        'var(amount)': 132918.536184,
        'std(amount)': 364.57994,
        'max(amount)': 1600,
        'mean(amount)': 105.65789473684211,
        'median(amount)': 12,
        'min(amount)': 2.0,
        'sum(amount)': 2007.5,
        'sum(gps_latitude)': 624.089497667,
        'ratio(amount, gps_latitude)': 3.184639,
        'sum(risk_factor in ["low_risk"])': 18,
        'ratio(risk_factor in ["low_risk"], risk_factor in ["low_risk",'
        ' "medium_risk"])': 18.0 / 19,
        'ratio(risk_factor in ["low_risk"], 1)': 18.0 / 19,
        'count()': 19.0,
        'count(risk_factor in ["low_risk"])': 18.0,
        'argmax(submit_date)': 18.0,
        'newest(submit_date, amount)': 28.0,
        'pearson(gps_latitude, amount)': -0.67643,
    }

    GROUP_TO_RESULTS = {
        'food_type':
        pickle.load(
            open('tests/fixtures/good_eats_agg_group_food_type.pkl', 'rb')),
        'food_type,rating':
        pickle.load(
            open('tests/fixtures/good_eats_agg_group_food_type_rating.pkl',
                 'rb')),
    }

    def setUp(self):
        TestCalculator.setUp(self)
        self.calculations = [
            'pearson(gps_latitude, amount)',
            'var(amount)',
            'std(amount)',
            'max(amount)',
            'mean(amount)',
            'median(amount)',
            'min(amount)',
            'sum(amount)',
            'sum(gps_latitude)',
            'ratio(amount, gps_latitude)',
            'sum(risk_factor in ["low_risk"])',
            'ratio(risk_factor in ["low_risk"], risk_factor in ["low_risk",'
            ' "medium_risk"])',
            'ratio(risk_factor in ["low_risk"], 1)',
            'count(risk_factor in ["low_risk"])',
            'count()',
            'argmax(submit_date)',
            'newest(submit_date, amount)',
        ]
        self.expected_length = defaultdict(int)
        self.groups_list = None

    def _offset_for_formula(self, formula, num_columns):
        if formula[:4] in ['mean', 'rati']:
            num_columns += 2
        elif formula[:7] == 'pearson':
            num_columns += 1

        return num_columns

    def _get_initial_len(self, formula, groups_list):
        initial_len = 0 if self.group == '' else len(groups_list)
        return self._offset_for_formula(formula, initial_len)

    def _columns_per_aggregation(self, formula, initial_num_columns=1):
        return self._offset_for_formula(formula, initial_num_columns)

    def _calculations_to_results(self, formula, row):
        if self.group:
            res = self.GROUP_TO_RESULTS[self.group][formula]
            column = row[self.groups_list[0]] if len(self.groups_list) <= 1\
                else tuple([row[group] for group in self.groups_list])
            res = res[column]
            return res
        else:
            return self.AGGREGATION_RESULTS[formula]

    def _test_calculation_results(self, name, formula):
        self.expected_length[self.group] += self._columns_per_aggregation(
            formula)

        # retrieve linked dataset
        linked_dset = self.dataset.aggregated_dataset(self.group)
        self.assertFalse(linked_dset is None)
        linked_dframe = linked_dset.dframe()

        name = linked_dset.schema.labels_to_slugs[name]

        self.assertTrue(name in linked_dframe.columns)
        self.assertEqual(len(linked_dframe.columns),
                         self.expected_length[self.group])

        # test that the schema is up to date
        self.assertTrue(linked_dset.SCHEMA in linked_dset.record.keys())
        self.assertTrue(isinstance(linked_dset.schema, dict))
        schema = linked_dset.schema

        # test slugified column names
        column_names = [name]
        if self.groups_list:
            column_names.extend(self.groups_list)
        for column_name in column_names:
            self.assertTrue(column_name in schema.keys())

        for idx, row in linked_dframe.iterrows():
            result = np.float64(row[name])
            stored = self._calculations_to_results(formula, row)
            # np.nan != np.nan, continue if we have two nan values
            if np.isnan(result) and np.isnan(stored):
                continue
            msg = self._equal_msg(result, stored, formula)
            self.assertAlmostEqual(result, stored, self.places, msg)

    def _test_aggregation(self):
        if self.group:
            self.groups_list = self.dataset.split_groups(self.group)
            self.expected_length[self.group] += len(self.groups_list)
        else:
            self.group = ''

        self._test_calculator()

    def test_aggregation(self):
        self._test_aggregation()

    def test_aggregation_with_group(self):
        self.group = 'food_type'
        self._test_aggregation()

    def test_aggregation_with_multigroup(self):
        self.group = 'food_type,rating'
        self._test_aggregation()



import urllib2

from bamboo.controllers.abstract_controller import AbstractController
from bamboo.core.frame import NonUniqueJoinError
from bamboo.core.merge import merge_dataset_ids, MergeError
from bamboo.core.summary import ColumnTypeError
from bamboo.lib.exceptions import ArgumentError
from bamboo.lib.jsontools import safe_json_loads
from bamboo.lib.utils import parse_int
from bamboo.lib.query_args import QueryArgs
from bamboo.models.dataset import Dataset
from bamboo.models.observation import Observation


class Datasets(AbstractController):
    """
    The Datasets Controller provides access to data.  Datasets can store data
    from uploaded CSVs or URLs pointing to CSVs.  Additional rows can be passed
    as JSON and added to a dataset.

    All actions in the Datasets Controller can optionally take a `callback`
    parameter.  If passed the returned result will be wrapped this the
    parameter value.  E.g., is ``callback=parseResults`` the returned value
    will be ``parseResults([some-JSON])``, where ``some-JSON`` is the function
    return value.

    Attributes:

    - SELECT_ALL_FOR_SUMMARY: a string the the client can pass to
    - indicate that all columns should be summarized.

    """
    SELECT_ALL_FOR_SUMMARY = 'all'

    def delete(self, dataset_id):
        """Delete the dataset with hash `dataset_id`.

        This also deletes any observations associated with the dataset_id. The
        dataset and observations are not recoverable.

        :param dataset_id: The dataset ID of the dataset to be deleted.

        :returns: A string of success or error if that dataset could not be
            found.
        """
        def action(dataset):
            dataset.delete()
            return {self.SUCCESS: 'deleted dataset: %s' % dataset_id}

        return self._safe_get_and_call(dataset_id, action)

    def info(self, dataset_id, callback=False):
        """Fetch and return the meta-data for a dataset.

        :param dataset_id: The dataset ID of the dataset to return meta-data
            for.
        :param callback: A JSONP callback function to wrap the result in.

        :returns: The data for `dataset_id`. Returns an error message if
            `dataset_id` does not exist.
        """
        def action(dataset):
            return dataset.info()

        return self._safe_get_and_call(dataset_id, action, callback=callback)

    def set_info(self, dataset_id, **kwargs):
        """Set the metadata for a dataset.

        :param dataset_id: ID of the dataset to update.
        :param attribution: Text to set the attribution to.
        :param description: Text to set the description to.
        :param label: Text to set the label to.
        :param license: Text to set the license to.

        :returns: Success or error.
        """
        def action(dataset):
            return dataset.info(kwargs)

        return self._safe_get_and_call(dataset_id, action)

    def summary(self, dataset_id, query={}, select=None,
                group=None, limit=0, order_by=None, callback=False):
        """Return a summary of the dataset ID given the passed parameters.

        Retrieve the dataset by ID then limit that data using the optional
        `query`, `select` and `limit` parameters. Summarize the resulting data
        potentially grouping by the optional `group` parameter. Order the
        results using `order_by` if passed.

        :param dataset_id: The dataset ID of the dataset to summarize.
        :param select: This is a required argument, it can be 'all' or a
            MongoDB JSON query.
        :param group: If passed, group the summary by this column or list of
            columns.
        :param query: If passed restrict summary to rows matching this query.
        :param limit: If passed limit the rows to summarize to this number.
        :param order_by: If passed order the result using this column.
        :param callback: A JSONP callback function to wrap the result in.

        :returns: An error message if `dataset_id` does not exist or the JSON
            for query or select is improperly formatted.  Otherwise the summary
            as a JSON string.

        :raises: `ArgumentError` if no select is supplied or dataset is not in
            ready state.
        """
        def action(dataset, query=query, select=select, limit=limit):
            if not dataset.is_ready:
                raise ArgumentError('dataset is not finished importing')

            limit = parse_int(limit, 0)
            query = self.__parse_query(query)
            select = self.__parse_select(select, required=True)

            groups = dataset.split_groups(group)

            # if select append groups to select
            if select:
                select.update(dict(zip(groups, [1] * len(groups))))

            query_args = QueryArgs(
                query=query, select=select, limit=limit, order_by=order_by)
            dframe = dataset.dframe(query_args)

            return dataset.summarize(dframe, groups=groups,
                                     no_cache=query or select)

        return self._safe_get_and_call(dataset_id, action, callback=callback,
                                       exceptions=(ColumnTypeError,))

    def aggregations(self, dataset_id, callback=False):
        """Return a dict of aggregated data for the given `dataset_id`.

        :param dataset_id: The dataset ID of the dataset to return aggregations
            for.
        :param callback: A JSONP callback function to wrap the result in.

        :returns: An error message if `dataset_id` does not exist. Otherwise,
            returns a dict of the form {[group]: [id]}.
        """
        def action(dataset):
            return dataset.aggregated_datasets_dict

        return self._safe_get_and_call(dataset_id, action, callback=callback)

    def show(self, dataset_id, query=None, select=None, distinct=None, limit=0,
             order_by=None, format=None, callback=False, count=False,
             index=False):
        """ Return rows for `dataset_id`, matching the passed parameters.

        Retrieve the dataset by ID then limit that data using the optional
        `query`, `select` and `limit` parameters. Order the results using
        `order_by` if passed.

        :param dataset_id: The dataset ID of the dataset to return.
        :param select: A MongoDB JSON query for select.
        :param distinct: A field to return distinct results for.
        :param query: If passed restrict results to rows matching this query.
        :param limit: If passed limit the rows to this number.
        :param order_by: If passed order the result using this column.
        :param format: Format of output data, 'json' or 'csv'
        :param callback: A JSONP callback function to wrap the result in.
        :param count: Return the count for this query.
        :param index: Include index with data.

        :returns: An error message if `dataset_id` does not exist or the JSON
            for query or select is improperly formatted. Otherwise a JSON
            string of the rows matching the parameters.
        """
        content_type = self.__content_type_for_format(format)

        def action(dataset, limit=limit, query=query, select=select):
            limit = parse_int(limit, 0)
            query = self.__parse_query(query)
            select = self.__parse_select(select)

            query_args = QueryArgs(
                query=query, select=select, distinct=distinct,
                limit=limit, order_by=order_by)

            if count:
                return dataset.count(query_args)
            else:
                dframe = dataset.dframe(query_args, index=index)

                if distinct:
                    return sorted(dframe[0].tolist())

            return self.__dataframe_as_content_type(content_type, dframe)

        return self._safe_get_and_call(
            dataset_id, action, callback=callback, content_type=content_type)

    def merge(self, dataset_ids, mapping=None):
        """Merge the datasets with the dataset_ids in `datasets`.

        :param dataset_ids: A JSON encoded array of dataset IDs for existing
            datasets.
        :param mapping: An optional mapping from original column names to
            destination names.

        :returns: An error if the datasets could not be found or less than two
            dataset IDs were passed.  Otherwise, the ID of the new merged
            dataset created by combining the datasets provided as an argument.
        """

        def action(dataset, dataset_ids=dataset_ids, mapping=mapping):
            if mapping:
                mapping = safe_json_loads(mapping)

            dataset_ids = safe_json_loads(dataset_ids)
            dataset = merge_dataset_ids(dataset_ids, mapping)

            return {Dataset.ID: dataset.dataset_id}

        return self._safe_get_and_call(
            None, action, exceptions=(MergeError,), error = 'merge failed')

    def create(self, url=None, csv_file=None, json_file=None, schema=None,
               perish=0):
        """Create a dataset by URL, CSV or schema file.

        If `url` is provided, create a dataset by downloading a CSV from that
        URL. If `url` is not provided and `csv_file` is provided, create a
        dataset with the data in the passed `csv_file`. If both `url` and
        `csv_file` are provided, `csv_file` is ignored. If `schema` is
        supplied, an empty dataset is created with the associated column
        structure.

        .. note::

            The follow words are reserved and will be slugified by adding
            underscores (or multiple underscores to ensure uniqueness) if used
            as column names:

                - all
                - and
                - case
                - date
                - default
                - in
                - not
                - or
                - sum
                - years

        :param url: A URL to load a CSV file from. The URL must point to a CSV
            file.
        :param csv_file: An uploaded CSV file to read from.
        :param json_file: An uploaded JSON file to read from.
        :param schema: A SDF schema file (JSON)
        :param perish: Number of seconds after which to delete the dataset.

        :returns: An error message if `url`, `csv_file`, or `scehma` are not
            provided. An error message if an improperly formatted value raises
            a ValueError, e.g. an improperly formatted CSV file. An error
            message if the URL could not be loaded. Otherwise returns a JSON
            string with the dataset ID of the newly created dataset.  Note that
            the dataset will not be fully loaded until its state is set to
            ready.
        """
        result = None
        error = 'url, csv_file or schema required'

        try:
            if schema or url or csv_file or json_file:
                dataset = Dataset()
                dataset.save()

                if schema:
                    dataset.import_schema(schema)

                if url:
                    dataset.import_from_url(url)
                elif csv_file:
                    dataset.import_from_csv(csv_file)
                elif json_file:
                    dataset.import_from_json(json_file)

                result = {Dataset.ID: dataset.dataset_id}

            perish = parse_int(perish, None)
            if perish:
                dataset.delete(countdown=perish)
        except urllib2.URLError:
            error = 'could not load: %s' % url
        except IOError:
            error = 'could not get a filehandle for: %s' % csv_file

        self.set_response_params(result, success_status_code=201)
        return self._dump_or_error(result, error)

    def update(self, dataset_id, update):
        """Update the `dataset_id` with the new rows as JSON.

        :param dataset_id: The ID of the dataset to update.
        :param update: The JSON to update the dataset with.

        :returns: A JSON dict with the ID of the dataset updated, or with an
            error message.
        """
        def action(dataset, update=update):
            update = safe_json_loads(update)
            dataset.add_observations(update)

            return {Dataset.ID: dataset_id}

        return self._safe_get_and_call(
            dataset_id, action, exceptions=(NonUniqueJoinError,))

    def drop_columns(self, dataset_id, columns):
        """Drop columns in dataset.

        Removes all the `columns` from the dataset with ID `dataset_id`.

        :param dataset_id: The ID of the dataset to update.
        :param columns: An array of columns within the dataset.

        :returns: An error if any column is not in the dataset. Otherwise a
            success message.
        """
        def action(dataset):
            dataset.drop_columns(columns)

            return {self.SUCCESS: 'in dataset %s dropped columns: %s' %
                    (dataset.dataset_id, columns)}

        return self._safe_get_and_call(dataset_id, action)

    def join(self, dataset_id, other_dataset_id, on=None):
        """Join the columns from two existing datasets.

        The `on` column must exists in both dataset. The values in the `on`
        `on` column of the other dataset must be unique.

        `on` can either be a single string to join on the same column in both
        datasets, or a comman separated list of the name for the left hand side
        and then the name for the right hand side. E.g.:

            * on=food_type,foodtype

        Will join on the left hand `food_type` column and the right hand
        `foodtype` column.


        :param dataset_id: The left hand dataset to be joined onto.
        :param other_dataset_id: The right hand to join.
        :param on: A column to join on, this column must be unique in the other
          dataset.

        :returns: Success and merged dataset ID or error message.
        """
        def action(dataset):
            other_dataset = Dataset.find_one(other_dataset_id)

            if other_dataset.record:
                merged_dataset = dataset.join(other_dataset, on)

                return {
                    self.SUCCESS: 'joined dataset %s to %s on %s' % (
                        other_dataset_id, dataset.dataset_id, on),
                    Dataset.ID: merged_dataset.dataset_id,
                }

        return self._safe_get_and_call(
            dataset_id, action, exceptions=(KeyError, NonUniqueJoinError))

    def resample(self, dataset_id, date_column, interval, how='mean',
                 query={}, format=None):
        """Resample a dataset."""
        content_type = self.__content_type_for_format(format)

        def action(dataset, query=query):
            query = self.__parse_query(query)

            dframe = dataset.resample(date_column, interval, how, query=query)
            return self.__dataframe_as_content_type(content_type, dframe)

        return self._safe_get_and_call(dataset_id, action,
                                       exceptions=(TypeError,),
                                       content_type=content_type)

    def set_olap_type(self, dataset_id, column, olap_type):
        """Set the OLAP Type for this `column` of dataset.

        Only columns with an original OLAP Type of 'measure' can be modified.
        This includes columns with Simple Type integer, float, and datetime.

        :param dataset_id: The ID of the dataset to modify.
        :param column: The column to set the OLAP Type for.
        :param olap_type: The OLAP Type to set. Must be 'dimension' or
            'measure'.
        """

        def action(dataset):
            dataset.set_olap_type(column, olap_type)

            return {self.SUCCESS: 'set OLAP Type for column "%s" to "%s".' % (
                column, olap_type),
                Dataset.ID: dataset_id}

        return self._safe_get_and_call(dataset_id, action)

    def rolling(self, dataset_id, window, win_type='boxcar',
                format=None):
        """Calculate the rolling window over a dataset."""
        content_type = self.__content_type_for_format(format)

        def action(dataset):
            dframe = dataset.rolling(str(win_type), int(window))
            return self.__dataframe_as_content_type(content_type, dframe)

        return self._safe_get_and_call(dataset_id, action,
                                       content_type=content_type)

    def row_delete(self, dataset_id, index):
        """Delete a row from dataset by index.

        :param dataset_id: The dataset to modify.
        :param index: The index to delete in the dataset.
        """
        def action(dataset):
            dataset.delete_observation(parse_int(index))

            return {
                self.SUCCESS: 'Deleted row with index "%s".' % index,
                Dataset.ID: dataset_id}

        return self._safe_get_and_call(dataset_id, action)

    def row_show(self, dataset_id, index):
        """Show a row by index.

        :param dataset_id: The dataset to fetch a row from.
        :param index: The index of the row to fetch.
        """
        def action(dataset):
            row = Observation.find_one(dataset, parse_int(index))

            if row:
                return row.clean_record

        error_message = "No row exists at index %s" % index
        return self._safe_get_and_call(dataset_id, action, error=error_message)

    def row_update(self, dataset_id, index, data):
        """Update a row in dataset by index.

        Update the row in dataset identified by `index` by updating it with the
        JSON dict `data`.  If there is a column in the DataFrame for which
        there is not a corresponding key in `data` that column will not be
        modified.

        :param dataset_id: The dataset to modify.
        :param index: The index to update.
        :param data: A JSON dict to update the row with.
        """
        def action(dataset, data=data):
            data = safe_json_loads(data)
            Observation.update(dataset, int(index), data)

            return {
                self.SUCCESS: 'Updated row with index "%s".' % index,
                Dataset.ID: dataset_id}

        return self._safe_get_and_call(dataset_id, action)

    def __dataframe_as_content_type(self, content_type, dframe):
        if content_type == self.CSV:
            return dframe.to_csv_as_string()
        else:
            return dframe.to_jsondict()

    def __content_type_for_format(self, format):
        return self.CSV if format == self.CSV else self.JSON

    def __parse_select(self, select, required=False):
        if required and select is None:
            raise ArgumentError('no select')

        if select == self.SELECT_ALL_FOR_SUMMARY:
            select = None
        elif select is not None:
            select = safe_json_loads(select, error_title='select')

            if not isinstance(select, dict):
                raise ArgumentError(
                    'select argument must be a JSON dictionary, found: %s.' %
                    select)

        return select

    def __parse_query(self, query):
        return safe_json_loads(query, error_title='string') if query else {}



import traceback

from celery.task import Task, task

from bamboo.core.calculator import Calculator
from bamboo.core.frame import DATASET_ID
from bamboo.lib.async import call_async
from bamboo.lib.exceptions import ArgumentError
from bamboo.lib.query_args import QueryArgs
from bamboo.lib.schema_builder import make_unique
from bamboo.models.abstract_model import AbstractModel


class UniqueCalculationError(Exception):

    def __init__(self, name, current_names):
        current_names_str = '", "'.join(current_names)
        message = ('The calculation name "%s" is not unique for this dataset.'
                   ' Please choose a name that does not exists.  The current '
                   'names are: "%s"' % (name, current_names_str))

        Exception.__init__(self, message)


class DependencyError(Exception):
    pass


@task(ignore_result=True)
def delete_task(calculation, dataset, slug):
    """Background task to delete `calculation` and columns in its dataset.

    Args:

    :param calculation: Calculation to delete.
    :param dataset: Dataset for this calculation.

    """
    dframe = dataset.dframe(keep_parent_ids=True)
    del dframe[slug]
    dataset.replace_observations(dframe, overwrite=True)
    calculation.remove_dependencies()
    dataset.clear_summary_stats(column=slug)

    super(calculation.__class__, calculation).delete({
        DATASET_ID: calculation.dataset_id,
        calculation.NAME: calculation.name
    })


class CalculateTask(Task):
    def after_return(self, status, retval, task_id, args, kwargs, einfo=None):
        if status == 'FAILURE':
            calculations = args[0]

            for calculation in calculations:
                calculation.failed(traceback.format_exc())


@task(base=CalculateTask, default_retry_delay=5, max_retries=10,
      ignore_result=True)
def calculate_task(calculations, dataset):
    """Background task to run a calculation.

    Set calculation to failed and raise if an exception occurs.

    :param calculation: Calculation to run.
    :param dataset: Dataset to run calculation on.
    """
    # block until other calculations for this dataset are finished
    calculations[0].restart_if_has_pending(dataset, calculations[1:])
    dataset.clear_summary_stats()

    calculator = Calculator(dataset)
    calculator.calculate_columns(calculations)

    for calculation in calculations:
        calculation.add_dependencies(dataset, calculator.dependent_columns())

        if calculation.aggregation is not None:
            aggregated_id = dataset.aggregated_datasets_dict[calculation.group]
            calculation.set_aggregation_id(aggregated_id)

        calculation.ready()


class Calculation(AbstractModel):

    __collectionname__ = 'calculations'

    AGGREGATION = 'aggregation'
    AGGREGATION_ID = 'aggregation_id'
    DEPENDENCIES = 'dependencies'
    DEPENDENT_CALCULATIONS = 'dependent_calculations'
    FORMULA = 'formula'
    GROUP = 'group'
    NAME = 'name'

    @property
    def aggregation(self):
        return self.record[self.AGGREGATION]

    @property
    def aggregation_id(self):
        return self.record.get(self.AGGREGATION_ID)

    @property
    def dataset_id(self):
        return self.record[DATASET_ID]

    @property
    def formula(self):
        return self.record[self.FORMULA]

    @property
    def group(self):
        return self.record[self.GROUP]

    @property
    def groups_as_list(self):
        return self.split_groups(self.group)

    @property
    def name(self):
        return self.record[self.NAME]

    @property
    def dependencies(self):
        return self.record.get(self.DEPENDENCIES, [])

    @property
    def dependent_calculations(self):
        return self.record.get(self.DEPENDENT_CALCULATIONS, [])

    def add_dependency(self, name):
        self.__add_and_update_set(self.DEPENDENCIES, self.dependencies, name)

    def add_dependent_calculation(self, name):
        self.__add_and_update_set(self.DEPENDENT_CALCULATIONS,
                                  self.dependent_calculations, name)

    def remove_dependent_calculation(self, name):
        new_dependent_calcs = self.dependent_calculations
        new_dependent_calcs.remove(name)
        self.update({self.DEPENDENT_CALCULATIONS: new_dependent_calcs})

    def remove_dependencies(self):
        for name in self.dependencies:
            calculation = self.find_one(self.dataset_id, name)
            calculation.remove_dependent_calculation(self.name)

    def set_aggregation_id(self, _id):
        self.update({self.AGGREGATION_ID: _id})

    def delete(self, dataset):
        """Delete this calculation.

        First ensure that there are no other calculations which depend on this
        one. If not, start a background task to delete the calculation.

        :param dataset: Dataset for this calculation.

        :raises: `DependencyError` if dependent calculations exist.
        :raises: `ArgumentError` if group is not in DataSet or calculation does
            not exist for DataSet.
        """
        if len(self.dependent_calculations):
            raise DependencyError(
                'Cannot delete, the calculations %s depend on this calculation'
                % self.dependent_calculations)

        if not self.group is None:
            # it is an aggregate calculation
            dataset = dataset.aggregated_dataset(self.group)

            if not dataset:
                raise ArgumentError(
                    'Aggregation with group "%s" does not exist for '
                    'dataset' % self.group)

        slug = dataset.schema.labels_to_slugs.get(self.name)

        if slug is None:
            raise ArgumentError(
                'Calculation "%s" does not exists for dataset' % self.name)

        call_async(delete_task, self, dataset, slug)

    def save(self, dataset, formula, name, group_str=None):
        """Parse, save, and calculate a formula.

        Validate `formula` and `group_str` for the given `dataset`. If the
        formula and group are valid for the dataset, then save a new
        calculation for them under `name`. Finally, create a background task
        to compute the calculation.

        Calculations are initially saved in a **pending** state, after the
        calculation has finished processing it will be in a **ready** state.

        :param dataset: The DataSet to save.
        :param formula: The formula to save.
        :param name: The name of the formula.
        :param group_str: Columns to group on.
        :type group_str: String, list or strings, or None.

        :raises: `ParseError` if an invalid formula was supplied.
        """
        calculator = Calculator(dataset)

        # ensure that the formula is parsable
        groups = self.split_groups(group_str) if group_str else []
        aggregation = calculator.validate(formula, groups)

        if aggregation:
            # set group if aggregation and group unset
            group_str = group_str or ''

            # check that name is unique for aggregation
            aggregated_dataset = dataset.aggregated_dataset(groups)

            if aggregated_dataset:
                name = self.__check_name_and_make_unique(name,
                                                         aggregated_dataset)

        else:
            name = self.__check_name_and_make_unique(name, dataset)

        record = {
            DATASET_ID: dataset.dataset_id,
            self.AGGREGATION: aggregation,
            self.FORMULA: formula,
            self.GROUP: group_str,
            self.NAME: name,
            self.STATE: self.STATE_PENDING,
        }
        super(self.__class__, self).save(record)

        return self

    @classmethod
    def create(cls, dataset, formula, name, group=None):
        calculation = super(cls, cls).create(dataset, formula, name, group)
        call_async(calculate_task, [calculation], dataset)
        return calculation

    @classmethod
    def create_from_list_or_dict(cls, dataset, calculations):
        if isinstance(calculations, dict):
            calculations = [calculations]

        if not len(calculations) or not isinstance(calculations, list):
            raise ArgumentError('Improper format for JSON calculations.')

        parsed_calculations = []

        # Pull out args to check JSON format
        try:
            for c in calculations:
                groups = c.get("groups")

                if not isinstance(groups, list):
                    groups = [groups]

                for group in groups:
                    parsed_calculations.append([
                        c[cls.FORMULA],
                        c[cls.NAME], group])

        except KeyError as e:
            raise ArgumentError('Required key %s not found in JSON' % e)

        calculations = [cls().save(dataset, formula, name, group)
                        for formula, name, group in parsed_calculations]
        call_async(calculate_task, calculations, dataset)

    @classmethod
    def find_one(cls, dataset_id, name, group=None):
        query = {
            DATASET_ID: dataset_id,
            cls.NAME: name,
        }

        if group:
            query[cls.GROUP] = group

        return super(cls, cls).find_one(query)

    @classmethod
    def find(cls, dataset):
        query_args = QueryArgs(query={DATASET_ID: dataset.dataset_id},
                               order_by='name')
        return super(cls, cls).find(query_args)

    def restart_if_has_pending(self, dataset, current_calcs=[]):
        current_names = sorted([self.name] + [c.name for c in current_calcs])
        unfinished = [c for c in dataset.calculations() if not c.is_ready]
        unfinished_names = [c.name for c in unfinished[:len(current_names)]]

        if len(unfinished) and current_names != sorted(unfinished_names):
            raise calculate_task.retry()

    def add_dependencies(self, dataset, dependent_columns):
        """Store calculation dependencies."""
        calculations = dataset.calculations()
        names_to_calcs = {calc.name: calc for calc in calculations}

        for column_name in dependent_columns:
            calc = names_to_calcs.get(column_name)
            if calc:
                self.add_dependency(calc.name)
                calc.add_dependent_calculation(self.name)

    def __check_name_and_make_unique(self, name, dataset):
        """Check that the name is valid and make unique if valid.

        :param name: The name to make unique.
        :param dataset: The dataset to make unique for.
        :raises: `UniqueCalculationError` if not unique.
        :returns: A unique name.
        """
        current_names = dataset.labels

        if name in current_names:
            raise UniqueCalculationError(name, current_names)

        return make_unique(name, dataset.schema.keys())

    def __add_and_update_set(self, link_key, existing, new):
        new_list = list(set(existing + [new]))

        if new_list != existing:
            self.update({link_key: new_list})



from math import isnan
from sys import maxint


def is_float_nan(num):
    """Return True is `num` is a float and NaN."""
    return isinstance(num, float) and isnan(num)


def minint():
    return -maxint - 1


def parse_int(value, default=None):
    return _parse_type(int, value, default)


def parse_float(value, default=None):
    return _parse_type(float, value, default)


def _parse_type(_type, value, default):
    try:
        return _type(value)
    except ValueError:
        return value if default is None else default
